"""
About the AudioMNIST corpus

The AudioMNIST dataset consists of 30000 audio recordings
(ca. 9.5 hours) of spoken digits (0-9) in English with 50 repetitions
per digit for each of the 60 different speakers. Recordings were
collected in quiet offices with a RØDE NT-USB microphone as
mono channel signal at a sampling frequency of 48kHz and were
saved in 16 bit integer format. In addition to audio recordings, meta
information including age (range: 22 - 61 years), gender (12 female
/ 48 male), origin and accent of all speakers were collected as well.
All speakers were informed about the intent of the data collection
and have given written declarations of consent for their participation prior
to their recording session.


The data is publicly available at the following github repo:

    https://github.com/soerenab/AudioMNIST
"""

import logging
import os
import tarfile
from pathlib import Path
from typing import Dict, Union

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.serialization import load_json
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download

# The last update is 5 years ago so it is safe to use the master branch.
AUDIO_MNIST_URL = "https://github.com/soerenab/AudioMNIST/archive/master.tar.gz"  # noqa


def download_audio_mnist(
    target_dir: Pathlike,
    force_download: bool = False,
) -> Path:
    """
    Download and untar the AudioMNIST dataset.

    :param target_dir: Pathlike, the path of the dir to store the dataset.
    :param force_download: bool, if True, download the archive even if it
      already exists.
    :return: Return the directory containing the extracted dataset.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    tgz_name = "master.tar.gz"
    tgz_path = target_dir / tgz_name
    if tgz_path.exists() and not force_download:
        logging.info(f"Skipping {tgz_name} because file exists.")

    resumable_download(
        AUDIO_MNIST_URL,
        tgz_path,
        force_download=force_download,
    )
    tgz_dir = target_dir / "AudioMNIST"
    if not tgz_dir.exists():
        logging.info(f"Untarring {tgz_name}.")
        with tarfile.open(tgz_path) as tar:
            tar.extractall(path=target_dir)
        os.rename(str(target_dir / "AudioMNIST-master"), str(tgz_dir))
    return tgz_dir


def prepare_audio_mnist(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """Prepare manifests for the AudioMNIST corpus.

    :param: corpus_dir: We assume it is the github repo directory and it
      contains the following directories: data/{01,02,03,...,60}
    :param: output_dir: Directory where the manifests should be written.
    """
    in_data_dir = Path(corpus_dir) / "data"
    assert (Path(in_data_dir) / "audioMNIST_meta.txt").is_file()

    metadata = load_json(in_data_dir / "audioMNIST_meta.txt")
    assert len(metadata) == 60, len(metadata)
    for i in range(1, 61):
        assert f"{i:02}" in metadata, i

    recordings = RecordingSet.from_recordings(
        Recording.from_file(
            path=path,
            # converts:
            #   path/to/AudioMNIST/data/01/3_01_24.wav
            # to:
            #   3_01_24
            recording_id=path.stem,
        )
        for i in range(1, 61)
        for path in (in_data_dir / f"{i:02}").rglob("*.wav")
    )

    id2text = {
        "0": "zero",
        "1": "one",
        "2": "two",
        "3": "three",
        "4": "four",
        "5": "five",
        "6": "six",
        "7": "seven",
        "8": "eight",
        "9": "nine",
    }

    supervisions = []
    for r in recordings:
        digit, speaker_id, _ = r.id.split("_")

        supervisions.append(
            SupervisionSegment(
                id=r.id,
                recording_id=r.id,
                start=0,
                duration=r.duration,
                channel=0,
                text=id2text[digit],
                language="English",
                speaker=speaker_id,
                custom=metadata[speaker_id],
            )
        )

    supervisions = SupervisionSet.from_segments(supervisions)
    recordings, supervisions = fix_manifests(recordings, supervisions)
    validate_recordings_and_supervisions(recordings, supervisions)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

        supervisions.to_file(output_dir / "audio_mnist_supervisions.jsonl.gz")
        recordings.to_file(output_dir / "audio_mnist_recordings.jsonl.gz")

    return {"recordings": recordings, "supervisions": supervisions}



"""
The Emilia dataset is constructed from a vast collection of speech data sourced
from diverse video platforms and podcasts on the Internet, covering various
content genres such as talk shows, interviews, debates, sports commentary, and
audiobooks. This variety ensures the dataset captures a wide array of real
human speaking styles. The initial version of the Emilia dataset includes a
total of 101,654 hours of multilingual speech data in six different languages:
English, French, German, Chinese, Japanese, and Korean.

See also
https://emilia-dataset.github.io/Emilia-Demo-Page/

Please note that Emilia does not own the copyright to the audio files; the
copyright remains with the original owners of the videos or audio. Users are
permitted to use this dataset only for non-commercial purposes under the
CC BY-NC-4.0 license.

Please refer to
https://huggingface.co/datasets/amphion/Emilia-Dataset
or
https://openxlab.org.cn/datasets/Amphion/Emilia
to download the dataset.

Note that you need to apply for downloading.

"""

from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Optional, Tuple

from tqdm.auto import tqdm

from lhotse import CutSet, MonoCut
from lhotse.audio import Recording
from lhotse.serialization import load_jsonl
from lhotse.supervision import SupervisionSegment
from lhotse.utils import Pathlike


def _parse_utterance(
    data_dir: Path,
    line: dict,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    """
    :param data_dir: Path to the data directory
    :param line: dict, it looks like below::

        {
          "id": "DE_B00000_S00000_W000029",
          "wav": "DE_B00000/DE_B00000_S00000/mp3/DE_B00000_S00000_W000029.mp3",
          "text": " Und es gibt auch einen Stadtplan von Tegun zu sehen.",
          "duration": 3.228,
          "speaker": "DE_B00000_S00000",
          "language": "de",
          "dnsmos": 3.3697
        }

    :return: a tuple of "recording" and "supervision"
    """
    full_path = data_dir / line["wav"]

    if not full_path.is_file():
        return None

    recording = Recording.from_file(
        path=full_path,
        recording_id=full_path.stem,
    )
    segment = SupervisionSegment(
        id=recording.id,
        recording_id=recording.id,
        start=0.0,
        duration=recording.duration,
        channel=0,
        text=line["text"],
        language=line["language"],
        speaker=line["speaker"],
        custom={"dnsmos": line["dnsmos"]},
    )

    return recording, segment


def prepare_emilia(
    corpus_dir: Pathlike,
    lang: str,
    num_jobs: int,
    output_dir: Optional[Pathlike] = None,
) -> CutSet:
    """
    Returns the manifests which consist of the Recordings and Supervisions

    :param corpus_dir: Pathlike, the path of the data dir.
                       We assume the directory has the following structure:
                       corpus_dir/raw/openemilia_all.tar.gz,
                       corpus_dir/raw/DE,
                       corpus_dir/raw/DE/DE_B00000.jsonl,
                       corpus_dir/raw/DE/DE_B00000/DE_B00000_S00000/mp3/DE_B00000_S00000_W000000.mp3,
                       corpus_dir/raw/EN, etc.
    :param lang: str, one of en, zh, de, ko, ja, fr
    :param num_jobs: int, number of threads for processing jsonl files
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: The CutSet containing the data for the given language.
    """
    if lang is None:
        raise ValueError("Please provide --lang")

    lang_uppercase = lang.upper()
    if lang_uppercase not in ("DE", "EN", "FR", "JA", "KO", "ZH"):
        raise ValueError(
            "Please provide a valid language. "
            f"Choose from de, en, fr, ja, ko, zh. Given: {lang}"
        )

    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    data_dir = corpus_dir / "raw" / lang_uppercase
    assert data_dir.is_dir(), f"No such directory: {data_dir}"

    jsonl_files = data_dir.glob("*.jsonl")

    cuts = []
    futures = []

    with ThreadPoolExecutor(num_jobs) as ex:
        for jsonl_file in jsonl_files:
            for item in tqdm(
                # Note: People's Speech manifest.json is really a JSONL.
                load_jsonl(jsonl_file),
                desc=f"Processing {jsonl_file} with {num_jobs} jobs",
            ):
                futures.append(
                    ex.submit(
                        _parse_utterance,
                        data_dir,
                        item,
                    )
                )

        for future in tqdm(futures, desc="Collecting futures"):
            result = future.result()
            if result is None:
                continue

            recording, segment = result

            cuts.append(
                MonoCut(
                    id=recording.id,
                    recording=recording,
                    start=0,
                    duration=recording.duration,
                    supervisions=[segment],
                    channel=0,
                )
            )

    cut_set = CutSet.from_cuts(cuts)

    if output_dir is not None:
        output_dir = Path(output_dir)
        cut_set.to_file(output_dir / f"emilia_cuts_{lang_uppercase}.jsonl.gz")

    return cut_set



from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterable, Optional, Sequence, Union

from lhotse import CutSet, FeatureSet, load_manifest
from lhotse.audio import RecordingSet
from lhotse.supervision import SupervisionSet
from lhotse.utils import Pathlike

DEFAULT_DETECTED_MANIFEST_TYPES = ("recordings", "supervisions")

TYPES_TO_CLASSES = {
    "recordings": RecordingSet,
    "supervisions": SupervisionSet,
    "features": FeatureSet,
    "cuts": CutSet,
}


def read_manifests_if_cached(
    dataset_parts: Optional[Sequence[str]],
    output_dir: Optional[Pathlike],
    prefix: str = "",
    suffix: Optional[str] = "jsonl.gz",
    types: Iterable[str] = DEFAULT_DETECTED_MANIFEST_TYPES,
    lazy: bool = False,
) -> Optional[Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]]:
    """
    Loads manifests from the disk, or a subset of them if only some exist.
    The manifests are searched for using the pattern ``output_dir / f'{prefix}_{manifest}_{part}.json'``,
    where `manifest` is one of ``["recordings", "supervisions"]`` and ``part`` is specified in ``dataset_parts``.
    This function is intended to speedup data preparation if it has already been done before.

    :param dataset_parts: Names of dataset pieces, e.g. in LibriSpeech: ``["test-clean", "dev-clean", ...]``.
    :param output_dir: Where to look for the files.
    :param prefix: Optional common prefix for the manifest files (underscore is automatically added).
    :param suffix: Optional common suffix for the manifest files ("json" by default).
    :param types: Which types of manifests are searched for (default: 'recordings' and 'supervisions').
    :return: A dict with manifest (``d[dataset_part]['recording'|'manifest']``) or ``None``.
    """
    if isinstance(dataset_parts, str):
        dataset_parts = [dataset_parts]
    if output_dir is None:
        return None
    if prefix and not prefix.endswith("_"):
        prefix = f"{prefix}_"
    if suffix.startswith("."):
        suffix = suffix[1:]
    if lazy and not suffix.startswith("jsonl"):
        raise ValueError(
            f"Only JSONL manifests can be opened lazily (got suffix: '{suffix}')"
        )
    manifests = defaultdict(dict)
    output_dir = Path(output_dir)
    for part in dataset_parts:
        for manifest in types:
            path = output_dir / f"{prefix}{manifest}_{part}.{suffix}"
            if not path.is_file():
                continue
            if lazy:
                manifests[part][manifest] = TYPES_TO_CLASSES[manifest].from_jsonl_lazy(
                    path
                )
            else:
                manifests[part][manifest] = load_manifest(path)
    return dict(manifests)


def manifests_exist(
    part: str,
    output_dir: Optional[Pathlike],
    types: Iterable[str] = DEFAULT_DETECTED_MANIFEST_TYPES,
    prefix: str = "",
    suffix: str = "jsonl.gz",
) -> bool:
    if output_dir is None:
        return False
    if prefix and not prefix.endswith("_"):
        prefix = f"{prefix}_"
    if suffix.startswith("."):
        suffix = suffix[1:]
    output_dir = Path(output_dir)
    for name in types:
        path = output_dir / f"{prefix}{name}_{part}.{suffix}"
        if not path.is_file():
            return False
    return True


def normalize_text_alimeeting(text: str, normalize: str = "m2met") -> str:
    """
    Text normalization similar to M2MeT challenge baseline.
    See: https://github.com/yufan-aslp/AliMeeting/blob/main/asr/local/text_normalize.pl
    """
    if normalize == "none":
        return text
    elif normalize == "m2met":
        import re

        text = text.replace("<sil>", "")
        text = text.replace("<%>", "")
        text = text.replace("<->", "")
        text = text.replace("<$>", "")
        text = text.replace("<#>", "")
        text = text.replace("<_>", "")
        text = text.replace("<space>", "")
        text = text.replace("`", "")
        text = text.replace("&", "")
        text = text.replace(",", "")
        if re.search("[a-zA-Z]", text):
            text = text.upper()
        text = text.replace("Ａ", "A")
        text = text.replace("ａ", "A")
        text = text.replace("ｂ", "B")
        text = text.replace("ｃ", "C")
        text = text.replace("ｋ", "K")
        text = text.replace("ｔ", "T")
        text = text.replace("，", "")
        text = text.replace("丶", "")
        text = text.replace("。", "")
        text = text.replace("、", "")
        text = text.replace("？", "")
        return text


def normalize_text_ami(text: str, normalize: str = "upper") -> str:
    """
    Text normalization similar to Kaldi's AMI recipe.
    """
    if normalize == "none":
        return text
    elif normalize == "upper":
        return text.upper()
    elif normalize == "kaldi":
        # Kaldi style text normalization
        import re

        # convert text to uppercase
        text = text.upper()
        # remove punctuations
        text = re.sub(r"[^A-Z0-9']+", " ", text)
        # remove multiple spaces
        text = re.sub(r"\s+", " ", text)
        # apply few exception for dashed phrases, Mm-Hmm, Uh-Huh, OK etc. those are frequent in AMI
        # and will be added to dictionary
        text = re.sub(r"MM HMM", "MM-HMM", text)
        text = re.sub(r"UH HUH", "UH-HUH", text)
        text = re.sub(r"(\b)O K(\b)", r"\g<1>OK\g<2>", text)
        text = re.sub(r"(\b)O_K(\b)", r"\g<1>OK\g<2>", text)
        return text.strip()


def normalize_text_chime6(text: str, normalize: str = "upper") -> str:
    """
    Text normalization similar to Kaldi's CHiME-6 recipe.
    """
    if normalize == "none":
        return text
    elif normalize == "upper":
        return text.upper()
    elif normalize == "kaldi":
        # Kaldi style text normalization
        import re

        if "[redacted]" in text:
            return ""

        # convert text to lowercase
        text = text.lower()
        # remove punctuations: " . ? , : ; !
        text = re.sub(r"[.?,:;!]", "", text)
        # remove multiple spaces
        text = re.sub(r"\s+", " ", text)
        # replace multiple consecutive [inaudible] with a single one
        text = re.sub(r"\[inaudible[- 0-9]*\]", "[inaudible]", text)
        # replace stranded dash with space
        text = re.sub(r" - ", " ", text)
        # replace mm- with mm
        text = re.sub(r"mm-", "mm", text)
        return text


def normalize_text_tedlium(text: str, normalize: str = "upper") -> str:
    """
    Text normalization similar to Kaldi's TEDLIUM-3 recipe.
    """
    if normalize == "none":
        return text
    elif normalize == "upper":
        return text.upper()
    elif normalize == "kaldi":
        # Kaldi style text normalization
        import re

        # remove tokens such as "[NOISE]"
        text = re.sub(r"\[[^\]]+\]", "", text)
        # remove "<unk>"
        text = re.sub(r"<unk>", "", text)
        # join suffixes with words, e.g. they 're -> they're
        text = re.sub(r"(\w+) '(\w+)", r"\1'\2", text)
        # join dangling "'" with next word, e.g. ' 60s -> '60s, ' cause -> 'cause
        text = re.sub(r"' (\w+)", r"'\1", text)

        return text.strip()


class TimeFormatConverter:
    @staticmethod
    def hms_to_seconds(time: str) -> float:
        """Converts time in HH:MM:SS.mmm format to seconds"""
        h, m, s = time.split(":")
        return int(h) * 3600 + int(m) * 60 + float(s)

    @staticmethod
    def seconds_to_hms(time: float) -> str:
        """Converts time in seconds to HH:MM:SS.mmm format"""
        h = int(time // 3600)
        m = int((time % 3600) // 60)
        s = time % 60
        return f"{h:02d}:{m:02d}:{s:06.3f}"



from typing import Optional, Sequence, Union

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.iwslt22_ta import prepare_iwslt22_ta
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("splits", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
@click.option(
    "--normalize-text",
    default=False,
    help="Whether to perform additional text cleaning and normalization from https://aclanthology.org/2022.iwslt-1.29.pdf.",
)
@click.option(
    "--langs",
    default="",
    help="Comma-separated list of language abbreviations for source and target languages",
)
def iwslt22_ta(
    corpus_dir: Pathlike,
    splits: Pathlike,
    output_dir: Pathlike,
    normalize_text: bool,
    langs: str,
    num_jobs: int,
):
    """
    IWSLT_2022 data preparation.
    \b
    This is conversational telephone speech collected as 8kHz-sampled data.
    The catalog number LDC2022E01 corresponds to the train, dev, and test1
    splits of the iwslt2022 shared task.
    To obtaining this data your institution needs to have an LDC subscription.
    You also should download the predined splits with
    git clone https://github.com/kevinduh/iwslt22-dialect.git
    """
    langs_list = langs.split(",")
    prepare_iwslt22_ta(
        corpus_dir,
        splits,
        output_dir=output_dir,
        num_jobs=num_jobs,
        clean=normalize_text,
        langs=langs_list,
    )



"""
Stcmds is an open-source  Chinese Mandarin corpus by Surfingtech (www.surfing.ai), containing utterances from 855 speakers, 102600 utterances;
Publicly available on https://www.openslr.org/resources/38
ST-CMDS (110 hours)

"""
import logging
import os
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download, safe_extract


def text_normalize(line: str):
    """
    Modified from https://github.com/wenet-e2e/wenet/blob/main/examples/multi_cn/s0/local/stcmds_data_prep.sh#L42
    paste -d' ' $data/utt.list $data/text.list |\
    sed 's/，//g' |\
    tr '[a-z]' '[A-Z]' |\
    awk '{if (NF > 1) print $0;}' > $data/train/text
    """
    line = line.replace("，", "")
    line = line.upper()
    return line


def download_stcmds(
    target_dir: Pathlike = ".",
    force_download: bool = False,
    base_url: str = "http://www.openslr.org/resources",
) -> Path:
    """
    Downdload and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    :return: the path to downloaded and extracted directory with data.
    """
    url = f"{base_url}/38"
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "stcmds"
    dataset_tar_name = "ST-CMDS-20170001_1-OS.tar.gz"
    for tar_name in [dataset_tar_name]:
        tar_path = target_dir / tar_name
        extracted_dir = corpus_dir / tar_name[:-7]
        completed_detector = extracted_dir / ".completed"
        if completed_detector.is_file():
            logging.info(f"Skipping download of because {completed_detector} exists.")
            continue
        resumable_download(
            f"{url}/{tar_name}", filename=tar_path, force_download=force_download
        )
        shutil.rmtree(extracted_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=corpus_dir)
        completed_detector.touch()

    return corpus_dir


def prepare_stcmds(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    path = corpus_dir / "ST-CMDS-20170001_1-OS"
    transcript_dict = {}
    for text_path in path.rglob("**/*.txt"):
        idx = text_path.stem
        logging.info(f"processing stcmds transcript  {text_path}")
        with open(text_path, "r", encoding="utf-8") as f:
            for line in f:
                line = text_normalize(line)
                transcript_dict[idx] = line

    manifests = defaultdict(dict)
    dataset_parts = ["train"]
    for part in tqdm(
        dataset_parts,
        desc="process stcmds audio, it needs waste about 2169 seconds time.",
    ):
        logging.info(f"Processing stcmds {part}")
        recordings = []
        supervisions = []
        for audio_path in path.rglob("**/*.wav"):
            idx = audio_path.stem
            speaker = "".join(list(idx)[8:15])
            if idx not in transcript_dict:
                logging.warning(f"No transcript: {idx}")
                logging.warning(f"{audio_path} has no transcript")
                continue
            text = transcript_dict[idx]
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            segment = SupervisionSegment(
                id=idx,
                recording_id=idx,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="Chinese",
                speaker=speaker,
                text=text.strip(),
            )
            supervisions.append(segment)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)

        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(output_dir / f"stcmds_supervisions_{part}.jsonl.gz")
            recording_set.to_file(output_dir / f"stcmds_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
Thchs is an open-source Chinese Speech Corpus Released by CSLT@Tsinghua University.
Publicly available on https://www.openslr.org/resources/18
THCHS-30 (26 hours)

"""


import logging
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download, safe_extract


def download_thchs_30(
    target_dir: Pathlike = ".",
    force_download: bool = False,
    base_url: str = "http://www.openslr.org/resources",
) -> Path:
    """
    Downdload and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    :return: the path to downloaded and extracted directory with data.
    """
    url = f"{base_url}/18"
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "thchs"
    dataset_tar_name = "data_thchs30.tgz"
    for tar_name in [dataset_tar_name]:
        tar_path = target_dir / tar_name
        extracted_dir = corpus_dir / tar_name[:-4]
        completed_detector = extracted_dir / ".completed"
        if completed_detector.is_file():
            logging.info(
                f"Skipping download {tar_name} because {completed_detector} exists."
            )
            continue
        resumable_download(
            f"{url}/{tar_name}", filename=tar_path, force_download=force_download
        )
        shutil.rmtree(extracted_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=corpus_dir)
        completed_detector.touch()

    return corpus_dir


"""
data_thchs30/data/B11_374.wav.trn
Its content is as follows:
徐 希君 肖 金生 刘 文华 屈 永利 王开 宇 骆 瑛 等 也 被 分别 判处 l = 六年 至 十 五年 有期徒刑
xu2 xi1 jun1 xiao1 jin1 sheng1 liu2 wen2 hua2 qu1 yong3 li4 wang2 kai1 yu3 luo4 ying1 deng3 ye3 bei4 fen1 bie2 pan4 chu3 liu4 nian2 zhi4 shi2 wu3 nian2 you3 qi1 tu2 xing2
x v2 x i1 j vn1 x iao1 j in1 sh eng1 l iu2 uu un2 h ua2 q v1 ii iong3 l i4 uu uang2 k ai1 vv v3 l uo4 ii ing1 d eng3 ii ie3 b ei4 f en1 b ie2 p an4 ch u3 l iu4 n ian2 zh ix4 sh ix2 uu u3 n ian2 ii iu3 q i1 t u2 x ing2
"""


def text_normalize(line: str):
    line = line.replace(" l =", "")
    line = line.upper()
    return line


def prepare_thchs_30(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    path = corpus_dir / "data_thchs30" / "data"
    transcript_dict = {}
    for text_path in path.rglob("**/*.wav.trn"):
        idx = Path(text_path.stem).stem
        with open(text_path, "r", encoding="utf-8") as f:
            for line_idx, line in enumerate(f):
                if line_idx == 0:
                    line = text_normalize(line)
                    transcript_dict[idx] = line
                continue

    manifests = defaultdict(dict)
    dataset_parts = ["train", "dev", "test"]
    for part in tqdm(
        dataset_parts,
        desc="Process thchs_30 audio, it takes about 19 seconds.",
    ):
        logging.info(f"Processing thchs_30 subset: {part}")
        # Generate a mapping: utt_id -> (audio_path, audio_info, speaker, text)
        recordings = []
        supervisions = []
        wav_path = corpus_dir / "data_thchs30" / f"{part}"
        for audio_path in wav_path.rglob("**/*.wav"):
            # logging.info(f"Processing audio path {audio_path}")
            idx = audio_path.stem
            speaker = idx.split("_")[0]
            if idx not in transcript_dict:
                logging.warning(f"No transcript: {idx}")
                logging.warning(f"{audio_path} has no transcript.")
                continue
            text = transcript_dict[idx]
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            segment = SupervisionSegment(
                id=idx,
                recording_id=idx,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="Chinese",
                speaker=speaker,
                text=text.strip(),
            )
            supervisions.append(segment)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)

        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"thchs_30_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(output_dir / f"thchs_30_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
The KeSpeech is an open source speech dataset, KeSpeech, which involves 1,542 hours of speech
signals recorded by 27,237 speakers in 34 cities in China, and the pronunciation includes
standard Mandarin and its 8 subdialects. The new dataset possesses several properties.
The dataset provides multiple labels including content transcription, speaker identity and
subdialect, hence supporting a variety of speech processing tasks, such as speech recognition,
speaker recognition, and subdialect identification, as well as other advanced techniques
like multi-task learning and conditional learning.

Full paper: https://openreview.net/forum?id=b3Zoeq2sCLq
"""

import logging
from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse.audio import AudioSource, Recording, RecordingSet, info
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.recipes.utils import manifests_exist
from lhotse.serialization import load_jsonl
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, compute_num_samples

KE_SPEECH_PARTS = ("train_phase1", "train_phase2", "dev_phase1", "dev_phase2", "test")


def text_normalize(line: str) -> str:
    line = line.replace("<SPOKEN_NOISE>", "")
    return line


def prepare_kespeech(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike],
    dataset_parts: Union[str, Sequence[str]] = "all",
    num_jobs: int = 1,
):
    corpus_dir = Path(corpus_dir)
    tasks_dir = corpus_dir / "Tasks" / "ASR"

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    assert tasks_dir.is_dir(), f"No such directory: {tasks_dir}"

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    subsets = KE_SPEECH_PARTS if "all" in dataset_parts else dataset_parts
    manifests = defaultdict(dict)
    for sub in subsets:
        if sub not in KE_SPEECH_PARTS:
            raise ValueError(f"No such part of dataset in KeSpeech : {sub}")
        manifests[sub] = {"recordings": [], "supervisions": []}

    with ThreadPoolExecutor(num_jobs) as ex:
        for part in tqdm(subsets, desc="Processing KeSpeech", unit="subset"):
            logging.info(f"Processing KeSpeech subset: {part}")

            if manifests_exist(part=part, output_dir=output_dir):
                logging.info(f"KeSpeech subset: {part} already prepared - skipping.")
                continue

            recordings = []
            supervisions = []
            part_path = tasks_dir / part
            futures = []

            with open(file=part_path / "wav.scp") as wav_scp, open(
                file=part_path / "text"
            ) as text, open(file=part_path / "utt2subdialect") as utt2subdialect, open(
                file=part_path / "utt2spk"
            ) as utt2spk:
                for wav_line, text_line, dialect_line, spk_line in zip(
                    wav_scp, text, utt2subdialect, utt2spk
                ):
                    futures.append(
                        ex.submit(
                            parse_utterance,
                            corpus_dir,
                            wav_line,
                            text_line,
                            dialect_line,
                            spk_line,
                        )
                    )
            for future in tqdm(futures, desc="Processing", leave=False):
                result = future.result()
                if result is None:
                    continue
                recording, segment = result
                recordings.append(recording)
                supervisions.append(segment)

            recording_set = RecordingSet.from_recordings(recordings)
            supervision_set = SupervisionSet.from_segments(supervisions)

            # Fix manifests
            recording_set, supervision_set = fix_manifests(
                recording_set, supervision_set
            )
            validate_recordings_and_supervisions(recording_set, supervision_set)

            if output_dir is not None:
                supervision_set.to_file(
                    output_dir / f"kespeech-asr_supervisions_{part}.jsonl.gz"
                )
                recording_set.to_file(
                    output_dir / f"kespeech-asr_recordings_{part}.jsonl.gz"
                )

            manifests[part] = {
                "recordings": recording_set,
                "supervisions": supervision_set,
            }
    return manifests


def parse_utterance(
    corpus_dir: Path,
    wav_line: str,
    text_line: str,
    dialect_line: str,
    spk_line: str,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    wav_id, wav_path = wav_line.strip().split(maxsplit=1)
    t_wav_id, transcript = text_line.strip().split(maxsplit=1)
    d_wav_id, dialect = dialect_line.strip().split(maxsplit=1)
    s_wav_id, speaker = spk_line.strip().split(maxsplit=1)
    assert wav_id == t_wav_id and t_wav_id == d_wav_id and d_wav_id == s_wav_id
    sampling_rate = 16000
    recording_info = info(corpus_dir / wav_path)
    recording = Recording(
        id=wav_id,
        sources=[
            AudioSource(
                type="file",
                channels=[0],
                source=str(corpus_dir / wav_path),
            )
        ],
        sampling_rate=sampling_rate,
        num_samples=compute_num_samples(
            duration=recording_info.duration,
            sampling_rate=sampling_rate,
        ),
        duration=recording_info.duration,
    )
    supervision = SupervisionSegment(
        id=wav_id,
        recording_id=wav_id,
        start=0.0,
        duration=recording_info.duration,
        text=text_normalize(transcript.strip()),
        language=dialect,
        speaker=speaker,
    )
    return recording, supervision



import warnings
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Union

import numpy as np
import torch

from lhotse.features.base import FeatureExtractor, register_extractor
from lhotse.features.kaldi.layers import (
    Wav2LogFilterBank,
    Wav2LogSpec,
    Wav2MFCC,
    Wav2Spec,
)
from lhotse.utils import (
    EPSILON,
    Seconds,
    asdict_nonull,
    compute_num_frames_from_samples,
)


@dataclass
class FbankConfig:
    sampling_rate: int = 16000
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    round_to_power_of_two: bool = True
    remove_dc_offset: bool = True
    preemph_coeff: float = 0.97
    window_type: str = "povey"
    dither: float = 0.0
    snip_edges: bool = False
    energy_floor: float = EPSILON
    raw_energy: bool = True
    use_energy: bool = False
    use_fft_mag: bool = False
    low_freq: float = 20.0
    high_freq: float = -400.0
    num_filters: int = 80
    num_mel_bins: Optional[int] = None  # do not use
    norm_filters: bool = False
    torchaudio_compatible_mel_scale: bool = True
    device: str = "cpu"

    def __post_init__(self):
        # This is to help users transition to a different Fbank implementation
        # from torchaudio.compliance.kaldi.fbank(), where the arg had a different name.
        if self.num_mel_bins is not None:
            self.num_filters = self.num_mel_bins
            self.num_mel_bins = None

        if self.snip_edges:
            warnings.warn(
                "`snip_edges` is set to True, which may cause issues in duration to num-frames conversion in Lhotse."
            )

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "FbankConfig":
        return FbankConfig(**data)


@register_extractor
class Fbank(FeatureExtractor):
    name = "kaldi-fbank"
    config_type = FbankConfig

    def __init__(self, config: Optional[FbankConfig] = None):
        super().__init__(config=config)
        config_dict = self.config.to_dict()
        config_dict.pop("device")
        self.extractor = Wav2LogFilterBank(**config_dict).to(self.device).eval()

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    def to(self, device: str):
        self.config.device = device
        self.extractor.to(device)

    def feature_dim(self, sampling_rate: int) -> int:
        return self.config.num_filters

    def extract(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> Union[np.ndarray, torch.Tensor]:
        assert sampling_rate == self.config.sampling_rate, (
            f"Fbank was instantiated for sampling_rate "
            f"{self.config.sampling_rate}, but "
            f"sampling_rate={sampling_rate} was passed to extract(). "
            "Note you can use CutSet/RecordingSet.resample() to change the audio sampling rate."
        )

        is_numpy = False
        if not isinstance(samples, torch.Tensor):
            samples = torch.from_numpy(samples)
            is_numpy = True

        if samples.ndim == 1:
            samples = samples.unsqueeze(0)

        feats = self.extractor(samples.to(self.device))[0]

        if is_numpy:
            return feats.cpu().numpy()
        else:
            return feats

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        return _extract_batch(
            self.extractor,
            samples,
            sampling_rate,
            frame_shift=self.frame_shift,
            lengths=lengths,
            device=self.device,
        )

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        return np.log(
            np.maximum(
                # protection against log(0); max with EPSILON is adequate since these are energies (always >= 0)
                EPSILON,
                np.exp(features_a) + energy_scaling_factor_b * np.exp(features_b),
            )
        )

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(np.exp(features)))


@dataclass
class MfccConfig:
    sampling_rate: int = 16000
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    round_to_power_of_two: bool = True
    remove_dc_offset: bool = True
    preemph_coeff: float = 0.97
    window_type: str = "povey"
    dither: float = 0.0
    snip_edges: bool = False
    energy_floor: float = EPSILON
    raw_energy: bool = True
    use_energy: bool = False
    use_fft_mag: bool = False
    low_freq: float = 20.0
    high_freq: float = -400.0
    num_filters: int = 23
    torchaudio_compatible_mel_scale: bool = True
    num_mel_bins: Optional[int] = None  # do not use
    norm_filters: bool = False
    num_ceps: int = 13
    cepstral_lifter: int = 22
    device: str = "cpu"

    def __post_init__(self):
        # This is to help users transition to a different Mfcc implementation
        # from torchaudio.compliance.kaldi.fbank(), where the arg had a different name.
        if self.num_mel_bins is not None:
            self.num_filters = self.num_mel_bins
            self.num_mel_bins = None

        if self.snip_edges:
            warnings.warn(
                "`snip_edges` is set to True, which may cause issues in duration to num-frames conversion in Lhotse."
            )

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "MfccConfig":
        return MfccConfig(**data)


@register_extractor
class Mfcc(FeatureExtractor):
    name = "kaldi-mfcc"
    config_type = MfccConfig

    def __init__(self, config: Optional[MfccConfig] = None):
        super().__init__(config=config)
        config_dict = self.config.to_dict()
        config_dict.pop("device")
        self.extractor = Wav2MFCC(**config_dict).to(self.device).eval()

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    def feature_dim(self, sampling_rate: int) -> int:
        return self.config.num_ceps

    def extract(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> Union[np.ndarray, torch.Tensor]:
        assert sampling_rate == self.config.sampling_rate, (
            f"Mfcc was instantiated for sampling_rate "
            f"{self.config.sampling_rate}, but "
            f"sampling_rate={sampling_rate} was passed to extract(). "
            "Note you can use CutSet/RecordingSet.resample() to change the audio sampling rate."
        )

        is_numpy = False
        if not isinstance(samples, torch.Tensor):
            samples = torch.from_numpy(samples)
            is_numpy = True

        if samples.ndim == 1:
            samples = samples.unsqueeze(0)

        feats = self.extractor(samples.to(self.device))[0]

        if is_numpy:
            return feats.cpu().numpy()
        else:
            return feats

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        return _extract_batch(
            self.extractor,
            samples,
            sampling_rate,
            frame_shift=self.frame_shift,
            lengths=lengths,
            device=self.device,
        )


@dataclass
class SpectrogramConfig:
    sampling_rate: int = 16000
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    round_to_power_of_two: bool = True
    remove_dc_offset: bool = True
    preemph_coeff: float = 0.97
    window_type: str = "povey"
    dither: float = 0.0
    snip_edges: bool = False
    energy_floor: float = EPSILON
    raw_energy: bool = True
    use_energy: bool = False
    use_fft_mag: bool = False
    device: str = "cpu"

    def __post_init__(self):
        if self.snip_edges:
            warnings.warn(
                "`snip_edges` is set to True, which may cause issues in duration to num-frames conversion in Lhotse."
            )

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "SpectrogramConfig":
        return SpectrogramConfig(**data)


@register_extractor
class Spectrogram(FeatureExtractor):
    name = "kaldi-spectrogram"
    config_type = SpectrogramConfig

    def __init__(self, config: Optional[SpectrogramConfig] = None):
        super().__init__(config=config)
        config_dict = self.config.to_dict()
        config_dict.pop("device")
        self.extractor = Wav2Spec(**config_dict).to(self.device).eval()

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    def feature_dim(self, sampling_rate: int) -> int:
        return self.extractor.fft_length // 2 + 1

    def extract(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> Union[np.ndarray, torch.Tensor]:
        assert sampling_rate == self.config.sampling_rate, (
            f"Spectrogram was instantiated for sampling_rate "
            f"{self.config.sampling_rate}, but "
            f"sampling_rate={sampling_rate} was passed to extract(). "
            "Note you can use CutSet/RecordingSet.resample() to change the audio sampling rate."
        )

        is_numpy = False
        if not isinstance(samples, torch.Tensor):
            samples = torch.from_numpy(samples)
            is_numpy = True

        if samples.ndim == 1:
            samples = samples.unsqueeze(0)

        feats = self.extractor(samples.to(self.device))[0]

        if is_numpy:
            return feats.cpu().numpy()
        else:
            return feats.cpu()

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        return _extract_batch(
            self.extractor,
            samples,
            sampling_rate,
            frame_shift=self.frame_shift,
            lengths=lengths,
            device=self.device,
        )

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        return features_a + energy_scaling_factor_b * features_b

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(features))


@dataclass
class LogSpectrogramConfig:
    sampling_rate: int = 16000
    frame_length: Seconds = 0.025
    frame_shift: Seconds = 0.01
    round_to_power_of_two: bool = True
    remove_dc_offset: bool = True
    preemph_coeff: float = 0.97
    window_type: str = "povey"
    dither: float = 0.0
    snip_edges: bool = False
    energy_floor: float = EPSILON
    raw_energy: bool = True
    use_energy: bool = False
    use_fft_mag: bool = False
    device: str = "cpu"

    def __post_init__(self):
        if self.snip_edges:
            warnings.warn(
                "`snip_edges` is set to True, which may cause issues in duration to num-frames conversion in Lhotse."
            )

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> "LogSpectrogramConfig":
        return LogSpectrogramConfig(**data)


@register_extractor
class LogSpectrogram(FeatureExtractor):
    name = "kaldi-log-spectrogram"
    config_type = LogSpectrogramConfig

    def __init__(self, config: Optional[LogSpectrogramConfig] = None):
        super().__init__(config=config)
        config_dict = self.config.to_dict()
        config_dict.pop("device")
        self.extractor = Wav2LogSpec(**config_dict).to(self.device).eval()

    @property
    def device(self) -> Union[str, torch.device]:
        return self.config.device

    @property
    def frame_shift(self) -> Seconds:
        return self.config.frame_shift

    def feature_dim(self, sampling_rate: int) -> int:
        return self.extractor.fft_length // 2 + 1

    def extract(
        self, samples: Union[np.ndarray, torch.Tensor], sampling_rate: int
    ) -> Union[np.ndarray, torch.Tensor]:
        assert sampling_rate == self.config.sampling_rate, (
            f"Spectrogram was instantiated for sampling_rate "
            f"{self.config.sampling_rate}, but "
            f"sampling_rate={sampling_rate} was passed to extract(). "
            "Note you can use CutSet/RecordingSet.resample() to change the audio sampling rate."
        )

        is_numpy = False
        if not isinstance(samples, torch.Tensor):
            samples = torch.from_numpy(samples)
            is_numpy = True

        if samples.ndim == 1:
            samples = samples.unsqueeze(0)

        feats = self.extractor(samples.to(self.device))[0]

        if is_numpy:
            return feats.cpu().numpy()
        else:
            return feats.cpu()

    def extract_batch(
        self,
        samples: Union[
            np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
        ],
        sampling_rate: int,
        lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    ) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
        return _extract_batch(
            self.extractor,
            samples,
            sampling_rate,
            frame_shift=self.frame_shift,
            lengths=lengths,
            device=self.device,
        )

    @staticmethod
    def mix(
        features_a: np.ndarray, features_b: np.ndarray, energy_scaling_factor_b: float
    ) -> np.ndarray:
        return features_a + energy_scaling_factor_b * features_b

    @staticmethod
    def compute_energy(features: np.ndarray) -> float:
        return float(np.sum(features))


def _extract_batch(
    extractor: FeatureExtractor,
    samples: Union[
        np.ndarray, torch.Tensor, Sequence[np.ndarray], Sequence[torch.Tensor]
    ],
    sampling_rate: int,
    frame_shift: Seconds = 0.01,
    lengths: Optional[Union[np.ndarray, torch.Tensor]] = None,
    device: Union[str, torch.device] = "cpu",
) -> Union[np.ndarray, torch.Tensor, List[np.ndarray], List[torch.Tensor]]:
    input_is_list = False
    input_is_torch = False

    if lengths is not None:
        feat_lens = [
            compute_num_frames_from_samples(l, frame_shift, sampling_rate)
            for l in lengths
        ]
        assert isinstance(
            samples, torch.Tensor
        ), "If `lengths` is provided, `samples` must be a batched and padded torch.Tensor."
    else:
        if isinstance(samples, list):
            input_is_list = True
            pass  # nothing to do with `samples`
        elif samples.ndim > 1:
            samples = list(samples)
        else:
            # The user passed an array/tensor of shape (num_samples,)
            samples = [samples.reshape(1, -1)]

        if any(isinstance(x, torch.Tensor) for x in samples):
            input_is_torch = True

        samples = [
            torch.from_numpy(x).squeeze() if isinstance(x, np.ndarray) else x.squeeze()
            for x in samples
        ]
        feat_lens = [
            compute_num_frames_from_samples(
                num_samples=len(x),
                frame_shift=extractor.frame_shift,
                sampling_rate=sampling_rate,
            )
            for x in samples
        ]
        samples = torch.nn.utils.rnn.pad_sequence(samples, batch_first=True)

    # Perform feature extraction
    input_device = samples.device
    feats = extractor(samples.to(device))
    feats.to(input_device)
    result = [feats[i, : feat_lens[i]] for i in range(len(samples))]

    if not input_is_torch:
        result = [x.numpy() for x in result]

    # If all items are of the same shape, concatenate
    if len(result) == 1:
        if input_is_list:
            return result
        else:
            return result[0]
    elif all(item.shape == result[0].shape for item in result[1:]):
        if input_is_torch:
            return torch.stack(result, dim=0)
        else:
            return np.stack(result, axis=0)
    else:
        return result



"""
The SpeechIO Chinese data is a collection of test sets covering wide range of speech recognition tasks & scenarios.

Participants can obtain the datasets at https://github.com/SpeechColab/Leaderboard - please download the datasets manually.
"""

import logging
import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.recipes.utils import manifests_exist
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available

SPEECHIO_TESTSET_INDEX = 26  # Currently, from 0 - 26 test sets are open source.


def _parse_one_subset(
    corpus_dir: Pathlike,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    recordings = []
    segments = []

    if not is_module_available("pandas"):
        raise ValueError("To prepare speechio data, please 'pip install pandas' first.")
    import pandas as pd

    df = pd.read_csv(f"{str(corpus_dir)}/metadata.tsv", sep="\t")

    recording_ids = df["ID"].tolist()
    texts = df["TEXT"].tolist()
    wav_paths = df["AUDIO"].tolist()

    for idx, audio_path in enumerate(wav_paths):
        audio_path = str(corpus_dir / audio_path)
        if not os.path.exists(audio_path):
            logging.warning(f"Audio file {audio_path} does not exist - skipping.")
            continue
        recording = Recording.from_file(audio_path)
        recordings.append(recording)
        recording_id = recording_ids[idx]
        text = texts[idx]
        speaker = recording_id.split("_")[0]

        segment = SupervisionSegment(
            id=f"{corpus_dir}-{recording_id}",
            recording_id=recording_id,
            start=0,
            duration=recording.duration,
            channel=0,
            language="Chinese",
            speaker=speaker,
            text=text,
        )
        segments.append(segment)

    return recordings, segments


def _prepare_subset(
    subset: str,
    corpus_dir: Pathlike,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param subset: str, the name of the subset.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / subset

    recording_set, supervision_set = _parse_one_subset(part_path)
    recording_set = RecordingSet.from_recordings(recording_set)
    supervision_set = SupervisionSet.from_segments(supervision_set)

    # Fix manifests
    recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
    validate_recordings_and_supervisions(recording_set, supervision_set)

    return recording_set, supervision_set


def prepare_speechio(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the SpeechIO dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing SpeechIO...")

    subsets = []
    for i in range(SPEECHIO_TESTSET_INDEX + 1):
        idx = f"{i}".zfill(2)
        subsets.append(f"SPEECHIO_ASR_ZH000{idx}")

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    for part in tqdm(subsets, desc="Dataset parts"):
        logging.info(f"Processing SpeechIO subset: {part}")
        if manifests_exist(
            part=part,
            output_dir=output_dir,
            prefix=f"speechio",
            suffix="jsonl.gz",
        ):
            logging.info(f"SpeechIO subset: {part} already prepared - skipping.")
            continue

        recording_set, supervision_set = _prepare_subset(part, corpus_dir)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"speechio_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(output_dir / f"speechio_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
Description taken from the abstract of the paper:
"EARS: An Anechoic Fullband Speech Dataset Benchmarked for Speech Enhancement and Dereverberation"
https://arxiv.org/abs/2406.06185

We release the EARS (Expressive Anechoic Recordings of Speech) dataset, a high-quality speech dataset comprising
107 speakers from diverse backgrounds, totaling in 100 hours of clean, anechoic speech data. The dataset covers
a large range of different speaking styles, including emotional speech, different reading styles, non-verbal sounds,
and conversational freeform speech. We benchmark various methods for speech enhancement and dereverberation on the
dataset and evaluate their performance through a set of instrumental metrics. In addition, we conduct a listening
test with 20 participants for the speech enhancement task, where a generative method is preferred. We introduce
a blind test set that allows for automatic online evaluation of uploaded data. Dataset download links and automatic
evaluation server can be found online.
"""


import json
import logging
import re
import shutil
import zipfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Iterable, Optional, Sequence, Union

from tqdm import tqdm

from lhotse import (
    RecordingSet,
    SupervisionSegment,
    SupervisionSet,
    fix_manifests,
    validate_recordings_and_supervisions,
)
from lhotse.recipes.utils import (
    DEFAULT_DETECTED_MANIFEST_TYPES,
    TYPES_TO_CLASSES,
    load_manifest,
    manifests_exist,
)
from lhotse.utils import Pathlike, resumable_download


def _read_manifests_if_cached_no_parts(
    output_dir: Optional[Pathlike],
    prefix: str = "",
    suffix: str = "jsonl.gz",
    types: Iterable[str] = DEFAULT_DETECTED_MANIFEST_TYPES,
    lazy: bool = False,
) -> Optional[Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]]:
    """
    Loads manifests from the disk, or a subset of them if only some exist.
    The manifests are searched for using the pattern ``output_dir / f'{prefix}_{manifest}_{part}.json'``,
    where `manifest` is one of ``["recordings", "supervisions"]`` and ``part`` is specified in ``dataset_parts``.
    This function is intended to speedup data preparation if it has already been done before.

    :param output_dir: Where to look for the files.
    :param prefix: Optional common prefix for the manifest files (underscore is automatically added).
    :param suffix: Optional common suffix for the manifest files ("json" by default).
    :param types: Which types of manifests are searched for (default: 'recordings' and 'supervisions').
    :return: A dict with manifest (``d[dataset_part]['recording'|'manifest']``) or ``None``.
    """
    if output_dir is None:
        return None
    if prefix and not prefix.endswith("_"):
        prefix = f"{prefix}_"
    if suffix.startswith("."):
        suffix = suffix[1:]
    if lazy and not suffix.startswith("jsonl"):
        raise ValueError(
            f"Only JSONL manifests can be opened lazily (got suffix: '{suffix}')"
        )
    manifests = defaultdict(dict)
    output_dir = Path(output_dir)
    for manifest in types:
        path = output_dir / f"{prefix}{manifest}.{suffix}"
        if not path.is_file():
            continue
        if lazy:
            manifests[manifest] = TYPES_TO_CLASSES[manifest].from_jsonl_lazy(path)
        else:
            manifests[manifest] = load_manifest(path)
    return dict(manifests)


def download_ears(
    target_dir: Pathlike = ".",
    force_download: bool = False,
) -> Path:
    """
    Download and unzip the EARS dataset.

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :return: the path to downloaded and extracted directory with data.
    """
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    resumable_download(
        "https://raw.githubusercontent.com/facebookresearch/ears_dataset/main/speaker_statistics.json",
        filename=target_dir / "speaker_statistics.json",
        force_download=force_download,
    )
    resumable_download(
        "https://raw.githubusercontent.com/facebookresearch/ears_dataset/main/transcripts.json",
        filename=target_dir / "transcripts.json",
        force_download=force_download,
    )
    for part in tqdm(
        range(1, 108), desc="Downloading the 107 speakers of the EARS dataset"
    ):
        part = f"p{part:03d}"
        url = f"https://github.com/facebookresearch/ears_dataset/releases/download/dataset"
        zip_name = f"{part}.zip"
        zip_path = target_dir / zip_name
        part_dir = target_dir / part
        completed_detector = part_dir / ".completed"
        if completed_detector.is_file():
            logging.info(f"Skipping {part} because {completed_detector} exists.")
            continue
        full_url = f"{url}/{zip_name}"
        resumable_download(full_url, filename=zip_path, force_download=force_download)
        shutil.rmtree(part_dir, ignore_errors=True)
        with zipfile.ZipFile(zip_path) as zf:
            zf.extractall(path=target_dir)
        completed_detector.touch()

    return target_dir


def prepare_ears(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param num_jobs: the number of parallel workers parsing the data.
    :return: a Dict whose keys are 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    dataset_parts = [f"p{spk:03d}" for spk in range(1, 108)]
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        # Maybe the manifests already exist: we can read them and save a bit of preparation time.
        manifests = _read_manifests_if_cached_no_parts(
            output_dir=output_dir, prefix="ears"
        )

    # Contents of the file
    # {
    #     "p001": {
    #         "age": "36-45",
    #         "ethnicity": "white or caucasian",
    #         "gender": "male",
    #         "weight": "160 - 180 lbs",
    #         "native language": "german",
    #         "height": "6' - 6'3"
    #     },
    #     ...
    # }
    spk2meta = json.loads((corpus_dir / "speaker_statistics.json").read_text())

    # Contents of the file
    # {
    #     "emo_adoration_sentences": "You're just the sweetest person I know and I am so happy to call you my friend. I had the best time with you, I just adore you. I love this gift, thank you!",
    #     "emo_amazement_sentences": "I just love how you can play guitar. You're so impressive. I admire your abilities so much.",
    #     ...
    # }
    utt2transcript = json.loads((corpus_dir / "transcripts.json").read_text())
    supervisions = []
    recordings_list = []
    for part in tqdm(dataset_parts, desc="Preparing EARS speakers"):
        if manifests_exist(part=part, output_dir=output_dir, prefix="ears"):
            logging.info(f"EARS subset: {part} already prepared - skipping.")
            continue
        spk_id = part
        part_path = corpus_dir / part
        recordings = RecordingSet.from_dir(
            part_path,
            "*.wav",
            num_jobs=num_jobs,
            recording_id=lambda path: f"{spk_id}_{path.stem}",
        )
        recordings_list.append(recordings)
        for rec in recordings:
            utt = rec.id.split("_")[1]
            meta = spk2meta[spk_id].copy()
            supervisions.append(
                SupervisionSegment(
                    id=rec.id,
                    recording_id=rec.id,
                    start=0.0,
                    duration=rec.duration,
                    channel=0,
                    text=utt2transcript.get(utt),
                    language="English",
                    speaker=spk_id,
                    gender=meta.pop("gender", None),
                    custom=meta,
                )
            )

    recordings = []
    for recs in recordings_list:
        recordings += list(recs)
    recordings = RecordingSet.from_recordings(recordings)
    supervisions = SupervisionSet.from_segments(supervisions)
    recordings, supervisions = fix_manifests(recordings, supervisions)
    validate_recordings_and_supervisions(recordings, supervisions)

    if output_dir is not None:
        supervisions.to_file(output_dir / f"ears_supervisions.jsonl.gz")
        recordings.to_file(output_dir / f"ears_recordings.jsonl.gz")

    manifests = {"recordings": recordings, "supervisions": supervisions}

    return manifests



#!/usr/bin/env python3
"""
Use this script like: https://lhotse.readthedocs.io/en/latest/cli.html
"""

# Note: we import all the CLI modes here so they get auto-registered
#       in Lhotse's main CLI entry-point. Then, setuptools is told to
#       invoke the "cli()" method from this script.
from lhotse.bin.modes import *



import functools
import logging
import os
import warnings
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Callable, Optional

from lhotse.utils import NonPositiveEnergyError, Seconds, fastcopy, suppress_and_warn

_DEFAULT_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE: Seconds = 0.025
_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE: Seconds = (
    _DEFAULT_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE
)


@dataclass
class VideoInfo:
    """
    Metadata about video content in a :class:`~lhotse.audio.Recording`.
    """

    fps: float
    """Video frame rate (frames per second). It's a float because some standard FPS are fractional (e.g. 59.94)"""

    num_frames: int
    """Number of video frames."""

    height: int
    """Height in pixels."""

    width: int
    """Width in pixels."""

    @property
    def duration(self) -> Seconds:
        return self.num_frames / self.fps

    @property
    def frame_length(self) -> Seconds:
        return 1.0 / self.fps

    def copy_with(self, **kwargs) -> "VideoInfo":
        return fastcopy(self, **kwargs)


def get_audio_duration_mismatch_tolerance() -> Seconds:
    """Retrieve the current audio duration mismatch tolerance in seconds."""
    if (
        _LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE
        != _DEFAULT_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE
    ):
        return _LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE

    if "LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE" in os.environ:
        return float(os.environ["LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE"])

    return _LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE


def set_audio_duration_mismatch_tolerance(delta: Seconds) -> None:
    """
    Override Lhotse's global threshold for allowed audio duration mismatch between the
    manifest and the actual data.

    Some scenarios when a mismatch can happen:

        - the :class:`.Recording` manifest duration is rounded off too much
            (i.e., bad user input, but too inconvenient to go back and fix the manifests)

        - data augmentation changes the number of samples a bit in a difficult to predict way

    When there is a mismatch, Lhotse will either trim or replicate the diff to match
    the value found in the :class:`.Recording` manifest.

    .. note:: We don't recommend setting this lower than the default value, as it could
        break some data augmentation transforms.

    Example::

        >>> import lhotse
        >>> lhotse.set_audio_duration_mismatch_tolerance(0.01)  # 10ms tolerance

    :param delta: New tolerance in seconds.
    """
    global _LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE
    logging.info(
        "The user overrided tolerance for audio duration mismatch "
        "between the values in the manifest and the actual data. "
        f"Old threshold: {_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE}s. "
        f"New threshold: {delta}s."
    )
    if delta < _DEFAULT_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE:
        warnings.warn(
            "The audio duration mismatch tolerance has been set to a value lower than "
            f"default ({_DEFAULT_LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE}s). "
            f"We don't recommend this as it might break some data augmentation transforms."
        )
    _LHOTSE_AUDIO_DURATION_MISMATCH_TOLERANCE = delta


class VideoLoadingError(Exception):
    pass


class AudioLoadingError(Exception):
    pass


class AudioSavingError(Exception):
    pass


class DurationMismatchError(Exception):
    pass


@contextmanager
def suppress_audio_loading_errors(enabled: bool = True):
    """
    Context manager that suppresses errors related to audio loading.
    Emits warning to the console.
    """
    with suppress_and_warn(
        AudioLoadingError,
        DurationMismatchError,
        NonPositiveEnergyError,
        ConnectionResetError,  # when reading from object stores / network sources
        enabled=enabled,
    ):
        yield


@contextmanager
def suppress_video_loading_errors(enabled: bool = True):
    """
    Context manager that suppresses errors related to audio loading.
    Emits warning to the console.
    """
    with suppress_and_warn(
        VideoLoadingError,
        AudioLoadingError,
        DurationMismatchError,
        NonPositiveEnergyError,
        ConnectionResetError,  # when reading from object stores / network sources
        enabled=enabled,
    ):
        yield


def null_result_on_audio_loading_error(func: Callable) -> Callable:
    """
    This is a decorator that makes a function return None when reading audio with Lhotse failed.

    Example::

        >>> @null_result_on_audio_loading_error
        ... def func_loading_audio(rec):
        ...     audio = rec.load_audio()  # if this fails, will return None instead
        ...     return other_func(audio)

    Another example::

        >>> # crashes on loading audio
        >>> audio = load_audio(cut)
        >>> # does not crash on loading audio, return None instead
        >>> maybe_audio: Optional = null_result_on_audio_loading_error(load_audio)(cut)
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> Optional:
        with suppress_audio_loading_errors():
            return func(*args, **kwargs)

    return wrapper


def verbose_audio_loading_exceptions() -> bool:
    return os.environ.get("LHOTSE_AUDIO_LOADING_EXCEPTION_VERBOSE") == "1"



from .audio import (
    AudioSource,
    Recording,
    RecordingSet,
    audio_backend,
    available_audio_backends,
    get_audio_duration_mismatch_tolerance,
    get_current_audio_backend,
    get_default_audio_backend,
    get_ffmpeg_torchaudio_info_enabled,
    set_audio_duration_mismatch_tolerance,
    set_current_audio_backend,
    set_ffmpeg_torchaudio_info_enabled,
)
from .caching import is_caching_enabled, set_caching_enabled
from .cut import CutSet, MonoCut, MultiCut, create_cut_set_eager, create_cut_set_lazy
from .features import *
from .kaldi import load_kaldi_data_dir
from .lazy import dill_enabled, is_dill_enabled, set_dill_enabled
from .manipulation import combine, split_parallelize_combine, to_manifest
from .qa import fix_manifests, validate, validate_recordings_and_supervisions
from .serialization import (
    available_io_backends,
    get_current_io_backend,
    get_default_io_backend,
    io_backend,
    load_manifest,
    load_manifest_lazy,
    set_current_io_backend,
    store_manifest,
)
from .supervision import SupervisionSegment, SupervisionSet
from .tools.env import add_tools_to_path as _add_tools_to_path
from .utils import (
    Decibels,
    Seconds,
    add_durations,
    compute_num_frames,
    compute_num_samples,
    fastcopy,
    fix_random_seed,
    measure_overlap,
    streaming_shuffle,
)
from .workflows import *

try:
    # Try to get Lhotse's version (should be created during running pip install / python setup.py ...)
    from .version import __version__
except:
    # Use a default placeholder when the version is unavailable...
    from os import environ as _environ
    from pathlib import Path as _Path

    _base_version_path = _Path(".").parent / "VERSION"
    if _base_version_path.is_file():
        _base_version = open(_base_version_path).read().strip()
        _dev_marker = ""
        if not _environ.get("LHOTSE_PREPARING_RELEASE", False):
            _dev_marker = ".dev"
        __version__ = f"{_base_version}{_dev_marker}+missing.version.file"
    else:
        __version__ = f"0.0.0+unknown.version"

from . import augmentation, dataset, features, recipes

_add_tools_to_path()



from functools import partial
from typing import Any, Dict, Optional

import numpy as np

from lhotse import Recording
from lhotse.utils import asdict_nonull, fastcopy, ifnone


class CustomFieldMixin:
    """
    :class:`CustomFieldMixin` is intended for classes such as Cut or SupervisionSegment
    that support holding custom, user-defined fields.

    .. caution:: Due to the way inheritance and dataclasses work before Python 3.10,
        it is necessary to re-define ``custom`` attribute in dataclasses that
        inherit from this mixin.
    """

    def __init__(self, custom: Optional[Dict[str, Any]]) -> None:
        self.custom: Optional[Dict[str, Any]] = custom

    def __setattr__(self, key: str, value: Any) -> None:
        """
        This magic function is called when the user tries to set an attribute.
        We use it as syntactic sugar to store custom attributes in ``self.custom``
        field, so that they can be (de)serialized later.
        Setting a ``None`` value will remove the attribute from ``custom``.
        """
        if key in self.__dataclass_fields__:
            super().__setattr__(key, value)
        else:
            custom = ifnone(self.custom, {})
            if value is None:
                custom.pop(key, None)
            else:
                custom[key] = value
            if custom:
                self.custom = custom

    def __getattr__(self, name: str) -> Any:
        """
        This magic function is called when the user tries to access an attribute
        of :class:`.MonoCut` that doesn't exist. It is used for accessing the custom
        attributes of cuts.

        We use it to look up the ``custom`` field: when it's None or empty,
        we'll just raise AttributeError as usual.
        If ``item`` is found in ``custom``, we'll return ``custom[item]``.
        If ``item`` starts with "load_", we'll assume the name of the relevant
        attribute comes after that, and that value of that field is of type
        :class:`~lhotse.array.Array` or :class:`~lhotse.array.TemporalArray`.
        We'll return its ``load`` method to call by the user.

        Example of attaching and reading an alignment as TemporalArray::

            >>> cut = MonoCut('cut1', start=0, duration=4, channel=0)
            >>> cut.alignment = TemporalArray(...)
            >>> ali = cut.load_alignment()

        """
        custom = self.custom
        if custom is None:
            raise AttributeError(f"No such attribute: {name}")
        if name in custom:
            # Somebody accesses raw [Temporal]Array manifest
            # or wrote a custom piece of metadata into MonoCut.
            return self.custom[name]
        elif name.startswith("load_"):
            # Return the method for loading [Temporal]Arrays,
            # to be invoked by the user.
            attr_name = name[5:]
            return partial(self.load_custom, attr_name)
        raise AttributeError(f"No such attribute: {name}")

    def __delattr__(self, key: str) -> None:
        """Used to support ``del cut.custom_attr`` syntax."""
        if key in self.__dataclass_fields__:
            super().__delattr__(key)
        if self.custom is None or key not in self.custom:
            raise AttributeError(f"No such member: '{key}'")
        del self.custom[key]

    def to_dict(self) -> Dict[str, Any]:
        return asdict_nonull(self)

    def with_custom(self, name: str, value: Any):
        """Return a copy of this object with an extra custom field assigned to it."""
        cpy = fastcopy(
            self, custom=self.custom.copy() if self.custom is not None else {}
        )
        cpy.custom[name] = value
        return cpy

    def load_custom(self, name: str) -> np.ndarray:
        """
        Load custom data as numpy array. The custom data is expected to have
        been stored in cuts ``custom`` field as an :class:`~lhotse.array.Array` or
        :class:`~lhotse.array.TemporalArray` manifest.

        .. note:: It works with Array manifests stored via attribute assignments,
            e.g.: ``cut.my_custom_data = Array(...)``.

        :param name: name of the custom attribute.
        :return: a numpy array with the data.
        """
        from lhotse.array import Array, TemporalArray

        value = self.custom.get(name)
        if isinstance(value, Array):
            # Array does not support slicing.
            return value.load()
        elif isinstance(value, TemporalArray):
            # TemporalArray supports slicing.
            return value.load(start=self.start, duration=self.duration)
        elif isinstance(value, Recording):
            # Recording supports slicing.
            # Note: cut.channels referes to cut.recording and not the custom field.
            # We have to use a special channel selector field instead; e.g.:
            # if this is "target_recording", we'll look for "target_recording_channel_selector"
            channels = self.custom.get(f"{name}_channel_selector")
            return value.load_audio(
                channels=channels, offset=self.start, duration=self.duration
            )
        else:
            raise ValueError(
                f"To load {name}, the cut needs to have field {name} (or cut.custom['{name}']) "
                f"defined, and its value has to be a manifest of type Array or TemporalArray."
            )

    def has_custom(self, name: str) -> bool:
        """
        Check if the Cut has a custom attribute with name ``name``.

        :param name: name of the custom attribute.
        :return: a boolean.
        """
        if self.custom is None:
            return False
        return name in self.custom

    def drop_custom(self, name: str):
        if self.custom is None or name not in self.custom:
            return None
        del self.custom[name]
        return self



import logging
import warnings
from dataclasses import dataclass
from functools import partial, reduce
from io import BytesIO
from operator import add
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

import numpy as np
import torch
from intervaltree import IntervalTree

from lhotse.array import Array, TemporalArray
from lhotse.audio import Recording, VideoInfo, get_audio_duration_mismatch_tolerance
from lhotse.audio.backend import save_audio
from lhotse.audio.mixer import AudioMixer, VideoMixer, audio_energy
from lhotse.augmentation import (
    AudioTransform,
    AugmentFn,
    LoudnessNormalization,
    ReverbWithImpulseResponse,
)
from lhotse.cut.base import Cut
from lhotse.cut.data import DataCut
from lhotse.cut.padding import PaddingCut
from lhotse.features import (
    FeatureExtractor,
    FeatureMixer,
    create_default_feature_extractor,
)
from lhotse.features.base import Features
from lhotse.features.io import FeaturesWriter
from lhotse.supervision import SupervisionSegment
from lhotse.utils import (
    DEFAULT_PADDING_VALUE,
    LOG_EPSILON,
    Decibels,
    Pathlike,
    Seconds,
    add_durations,
    compute_num_frames,
    compute_num_samples,
    fastcopy,
    hash_str_to_int,
    merge_items_with_delimiter,
    overlaps,
    perturb_num_samples,
    rich_exception_info,
    uuid4,
)


@dataclass
class MixTrack:
    """
    Represents a single track in a mix of Cuts. Points to a specific DataCut or PaddingCut and holds information on
    how to mix it with other Cuts, relative to the first track in a mix.
    """

    cut: Union[DataCut, PaddingCut]
    type: str = None
    offset: Seconds = 0.0
    snr: Optional[Decibels] = None

    def __post_init__(self):
        self.type = type(self.cut).__name__

    @staticmethod
    def from_dict(data: dict):
        from .set import deserialize_cut

        # Take out `type` from data dict and put it into the `cut` dict.
        cut_dict = data.pop("cut")
        cut_dict["type"] = data.pop("type")
        return MixTrack(deserialize_cut(cut_dict), **data)

    def to_dict(self) -> Dict:
        ans = {
            "cut": self.cut.to_dict(),
            "type": self.type,
            "offset": self.offset,
        }
        if self.snr is not None:
            ans["snr"] = self.snr
        return ans


@dataclass
class MixedCut(Cut):
    """
    :class:`~lhotse.cut.MixedCut` is a :class:`~lhotse.cut.Cut` that actually consists of multiple other cuts.
    Its primary purpose is to allow time-domain and feature-domain augmentation via mixing the training cuts
    with noise, music, and babble cuts. The actual mixing operations are performed on-the-fly.

    Internally, :class:`~lhotse.cut.MixedCut` holds other cuts in multiple tracks (:class:`~lhotse.cut.MixTrack`),
    each with its own offset and SNR that is relative to the first track.

    Please refer to the documentation of :class:`~lhotse.cut.Cut` to learn more about using cuts.

    In addition to methods available in :class:`~lhotse.cut.Cut`, :class:`~lhotse.cut.MixedCut` provides the methods to
    read all of its tracks audio and features as separate channels:

        >>> cut = MixedCut(...)
        >>> mono_features = cut.load_features()
        >>> assert len(mono_features.shape) == 2
        >>> multi_features = cut.load_features(mixed=False)
        >>> # Now, the first dimension is the channel.
        >>> assert len(multi_features.shape) == 3

    .. note:: MixedCut is different from MultiCut, which is intended to represent multi-channel recordings
        that share the same supervisions.

    .. note:: Each track in a MixedCut can be either a MonoCut, MultiCut, or PaddingCut.

    .. note:: The ``transforms`` field is a list of dictionaries that describe the transformations
        that should be applied to the track after mixing.

    See also:

        - :class:`lhotse.cut.Cut`
        - :class:`lhotse.cut.MonoCut`
        - :class:`lhotse.cut.MultiCut`
        - :class:`lhotse.cut.CutSet`
    """

    id: str
    tracks: List[MixTrack]
    transforms: Optional[List[AudioTransform]] = None

    @property
    def supervisions(self) -> List[SupervisionSegment]:
        """
        Lists the supervisions of the underlying source cuts.
        Each segment start time will be adjusted by the track offset.
        """
        return [
            segment.with_offset(track.offset)
            for track in self.tracks
            for segment in track.cut.supervisions
        ]

    @property
    def start(self) -> Seconds:
        return 0

    @property
    def duration(self) -> Seconds:
        track_durations = (track.offset + track.cut.duration for track in self.tracks)
        return round(max(track_durations), ndigits=8)

    @property
    def channel(self) -> Union[int, List[int]]:
        num_channels = self.num_channels
        return list(range(num_channels)) if num_channels > 1 else 0

    @property
    def has_features(self) -> bool:
        return self._first_non_padding_cut.has_features

    @property
    def has_recording(self) -> bool:
        return self._first_non_padding_cut.has_recording

    @property
    def has_video(self) -> bool:
        return self._first_non_padding_cut.has_video

    @property
    def is_in_memory(self) -> bool:
        return any(track.cut.is_in_memory for track in self.tracks)

    def has(self, field: str) -> bool:
        return self._first_non_padding_cut.has(field)

    @property
    def num_frames(self) -> Optional[int]:
        if self.has_features:
            return compute_num_frames(
                duration=self.duration,
                frame_shift=self.frame_shift,
                sampling_rate=self.sampling_rate,
            )
        return None

    @property
    def frame_shift(self) -> Optional[Seconds]:
        return self.tracks[0].cut.frame_shift

    @property
    def sampling_rate(self) -> Optional[int]:
        return self.tracks[0].cut.sampling_rate

    @property
    def num_samples(self) -> Optional[int]:
        return compute_num_samples(self.duration, self.sampling_rate)

    @property
    def num_features(self) -> Optional[int]:
        return self.tracks[0].cut.num_features

    @property
    def num_channels(self) -> Optional[int]:
        # PaddingCut and MonoCut have 1 channel each, MultiCut has N. We don't support MixedCut with MixedCut tracks.
        return max(track.cut.num_channels for track in self.tracks)

    @property
    def features_type(self) -> Optional[str]:
        return self._first_non_padding_cut.features.type if self.has_features else None

    def to_dict(self) -> dict:
        ans = {
            "id": self.id,
            "tracks": [t.to_dict() for t in self.tracks],
            "type": type(self).__name__,
        }
        if self.transforms:
            ans["transforms"] = [t.to_dict() for t in self.transforms]
        return ans

    def iter_data(
        self,
    ) -> Generator[
        Tuple[str, Union[Recording, Features, Array, TemporalArray]], None, None
    ]:
        """
        Iterate over each data piece attached to this cut.
        Returns a generator yielding tuples of ``(key, manifest)``, where
        ``key`` is the name of the attribute under which ``manifest`` is found.
        ``manifest`` is of type :class:`~lhotse.Recording`, :class:`~lhotse.Features`,
        :class:`~lhotse.TemporalArray`, or :class:`~lhotse.Array`.

        For example, if ``key`` is ``recording``, then ``manifest`` is ``self.recording``.
        """
        return self._first_non_padding_cut.iter_data()

    def __setattr__(self, key: str, value: Any) -> None:
        """
        This magic function is called when the user tries to set an attribute.
        We use it as syntactic sugar to store custom attributes in ``self._first_non_padding_cut.custom``
        field, so that they can be (de)serialized later.
        Setting a ``None`` value will remove the attribute from ``custom``.

        .. note:: MixedCut doesn't have its own ``custom`` field, and by convention
            always refers to the ``custom`` field on its first non padding cut.
        """
        if key in self.__dataclass_fields__:
            super().__setattr__(key, value)
        else:
            setattr(self._first_non_padding_cut, key, value)

    def __getattr__(self, name: str) -> Any:
        """
        This magic function is called when the user tries to access an attribute
        of :class:`.MixedCut` that doesn't exist. It is used for accessing the custom
        attributes of cuts. We support exactly one scenario for mixed cuts:

        If :attr:`tracks` contains exactly one :class:`.MonoCut` object (and an arbitrary
        number of :class:`.PaddingCut` objects), we will look up the custom attributes
        of that cut.

        If one of the custom attributes is of type :class:`~lhotse.array.Array` or
        :class:`~lhotse.array.TemporalArray` we'll also support loading those arrays
        (see example below). Additionally, we will incorporate extra padding as
        dictated by padding cuts.

        Example:

            >>> cut = MonoCut('cut1', start=0, duration=4, channel=0)
            >>> cut.alignment = TemporalArray(...)
            >>> mixed_cut = cut.pad(10, pad_value_dict={'alignment': -1})
            >>> ali = mixed_cut.load_alignment()

        """
        # Python will sometimes try to call undefined magic functions,
        # just fail for them (e.g. __setstate__ when pickling).
        if name.startswith("__"):
            raise AttributeError()

        # Loading a custom array attribute + performing padding.
        if name.startswith("load_"):
            attr_name = name[5:]
            return partial(self.load_custom, attr_name)

        # Returning the contents of "mono_cut.custom[name]",
        # or raising AttributeError.
        try:
            (
                non_padding_idx,
                mono_cut,
            ) = self._assert_one_data_cut_with_attr_and_return_it_with_track_index(name)
            return getattr(mono_cut, name)
        except AssertionError:
            raise AttributeError(
                f"No such attribute: '{name}' (note: custom attributes are not supported "
                f"when a MixedCut consists of more than one MonoCut with that attribute)."
            )

    def has_custom(self, name: str) -> bool:
        (
            non_padding_idx,
            mono_cut,
        ) = self._assert_one_data_cut_with_attr_and_return_it_with_track_index(name)

        return hasattr(mono_cut, name)

    def load_custom(self, name: str) -> np.ndarray:
        """
        Load custom data as numpy array. The custom data is expected to have
        been stored in cuts ``custom`` field as an :class:`~lhotse.array.Array` or
        :class:`~lhotse.array.TemporalArray` manifest.

        .. note:: It works with Array manifests stored via attribute assignments,
            e.g.: ``cut.my_custom_data = Array(...)``.

        .. warning:: For :class:`.MixedCut`, this will only work if the mixed cut
            consists of a single :class:`.MonoCut` and an arbitrary number of
            :class:`.PaddingCuts`. This is because it is generally undefined how to
            mix arbitrary arrays.

        :param name: name of the custom attribute.
        :return: a numpy array with the data (after padding).
        """

        from lhotse.array import Array, pad_array

        (
            non_padding_idx,
            mono_cut,
        ) = self._assert_one_data_cut_with_attr_and_return_it_with_track_index(name)

        # Use getattr to propagate AttributeError if "name" is not defined.
        manifest = getattr(mono_cut, name)

        # Check if the corresponding manifest for 'load_something' is of type
        # Array; if yes, just return the loaded data.
        # This is likely an embedding without a temporal dimension.
        if isinstance(manifest, Array):
            return mono_cut.load_custom(name)

        # We are loading either an array with a temporal dimension, or a recording:
        # We need to pad it.
        left_padding = self.tracks[non_padding_idx].offset
        padded_duration = self.duration

        # Then, check if it's a Recording. In that case we convert it to a cut,
        # leverage existing padding methods, and load padded audio data.
        if isinstance(manifest, Recording):
            return (
                manifest.to_cut()
                .pad(duration=manifest.duration + left_padding, direction="left")
                .pad(duration=padded_duration, direction="right")
                .load_audio()
            )

        # Load the array and retrieve the manifest from the only non-padding cut.
        # We'll also need to fetch the dict defining what padding value to use (if present).
        array = mono_cut.load_custom(name)
        try:
            pad_value_dict = [
                t.cut for t in self.tracks if isinstance(t.cut, PaddingCut)
            ][0].custom
            pad_value = pad_value_dict[name]
        except:
            pad_value = DEFAULT_PADDING_VALUE

        return pad_array(
            array,
            temporal_dim=manifest.temporal_dim,
            frame_shift=manifest.frame_shift,
            offset=left_padding,
            padded_duration=padded_duration,
            pad_value=pad_value,
        )

    def _assert_one_data_cut_with_attr_and_return_it_with_track_index(
        self,
        attr_name: str,
    ) -> Tuple[int, DataCut]:
        # TODO(pzelasko): consider relaxing this condition to
        #                 supporting mixed cuts that are not overlapping
        non_padding_cuts = [
            (idx, t.cut)
            for idx, t in enumerate(self.tracks)
            if isinstance(t.cut, DataCut)
        ]
        non_padding_cuts_with_custom_attr = [
            (idx, cut)
            for idx, cut in non_padding_cuts
            if cut.custom is not None and attr_name in cut.custom
        ]
        assert len(non_padding_cuts_with_custom_attr) == 1, (
            f"This MixedCut has {len(non_padding_cuts_with_custom_attr)} non-padding cuts "
            f"with a custom attribute '{attr_name}'. We currently don't support mixing custom attributes. "
            f"Consider dropping the attribute on all but one of DataCuts. Problematic cut:\n{self}"
        )
        non_padding_idx, mono_cut = non_padding_cuts_with_custom_attr[0]
        return non_padding_idx, mono_cut

    def move_to_memory(
        self,
        audio_format: str = "flac",
        load_audio: bool = True,
        load_features: bool = True,
        load_custom: bool = True,
    ) -> "MixedCut":
        """
        Load data (audio, features, or custom arrays) into memory and attach them
        to a copy of the manifest. This is useful when you want to store cuts together
        with the actual data in some binary format that enables sequential data reads.

        Audio is encoded with ``audio_format`` (compatible with ``torchaudio.save``),
        floating point features are encoded with lilcom, and other arrays are pickled.
        """
        return fastcopy(
            self,
            tracks=[
                fastcopy(
                    t,
                    cut=t.cut.move_to_memory(
                        audio_format=audio_format,
                        load_audio=load_audio,
                        load_features=load_features,
                        load_custom=load_custom,
                    ),
                )
                for t in self.tracks
            ],
        )

    def to_mono(
        self,
        encoding: str = "flac",
        **kwargs,
    ) -> "Cut":
        """
        Convert this MixedCut to a MonoCut by mixing all tracks and channels into a single one.
        The result audio array is stored in memory, and can be saved to disk by calling
        ``cut.save_audio(path, ...)`` on the result.

        .. hint:: the resulting MonoCut will have ``custom`` field populated with the
            ``custom`` value from the first track of the MixedCut.

        :param encoding: any of "wav", "flac", or "opus".
        :return: a new MonoCut instance.
        """
        samples = self.load_audio(mono_downmix=True)
        stream = BytesIO()
        save_audio(
            stream,
            samples,
            self.sampling_rate,
            format=encoding,
        )
        recording = Recording.from_bytes(stream.getvalue(), recording_id=self.id)
        return fastcopy(
            recording.to_cut(),
            supervisions=[fastcopy(s, channel=0) for s in self.supervisions],
            custom=self.tracks[0].cut.custom,
        )

    def truncate(
        self,
        *,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        keep_excessive_supervisions: bool = True,
        preserve_id: bool = False,
        _supervisions_index: Optional[Dict[str, IntervalTree]] = None,
    ) -> Cut:
        """
        Returns a new MixedCut that is a sub-region of the current MixedCut. This method truncates the underlying Cuts
        and modifies their offsets in the mix, as needed. Tracks that do not fit in the truncated cut are removed.

        Note that no operation is done on the actual features - it's only during the call to load_features()
        when the actual changes happen (a subset of features is loaded).

        :param offset: float (seconds), controls the start of the new cut relative to the current MixedCut's start.
        :param duration: optional float (seconds), controls the duration of the resulting MixedCut.
            By default, the duration is (end of the cut before truncation) - (offset).
        :param keep_excessive_supervisions: bool. Since trimming may happen inside a SupervisionSegment, the caller has
            an option to either keep or discard such supervisions.
        :param preserve_id: bool. Should the truncated cut keep the same ID or get a new, random one.
        :return: a new MixedCut instance.
        """

        assert (
            offset >= 0
        ), f"Offset for truncate must be non-negative (provided {offset})."
        new_tracks = []
        old_duration = self.duration
        new_mix_end = (
            add_durations(old_duration, -offset, sampling_rate=self.sampling_rate)
            if duration is None
            else add_durations(offset, duration, sampling_rate=self.sampling_rate)
        )

        for track in sorted(self.tracks, key=lambda t: t.offset):
            # First, determine how much of the beginning of the current track we're going to truncate:
            # when the track offset is larger than the truncation offset, we are not truncating the cut;
            # just decreasing the track offset.

            # 'cut_offset' determines how much we're going to truncate the Cut for the current track.
            cut_offset = max(
                add_durations(offset, -track.offset, sampling_rate=self.sampling_rate),
                0,
            )
            # 'track_offset' determines the new track's offset after truncation.
            track_offset = max(
                add_durations(track.offset, -offset, sampling_rate=self.sampling_rate),
                0,
            )
            # 'track_end' is expressed relative to the beginning of the mix
            # (not to be confused with the 'start' of the underlying MonoCut)
            track_end = add_durations(
                track.offset, track.cut.duration, sampling_rate=self.sampling_rate
            )

            if track_end < offset:
                # Omit a Cut that ends before the truncation offset.
                continue

            cut_duration_decrease = 0
            if track_end > new_mix_end:
                if duration is not None:
                    cut_duration_decrease = add_durations(
                        track_end, -new_mix_end, sampling_rate=self.sampling_rate
                    )
                else:
                    cut_duration_decrease = add_durations(
                        track_end, -old_duration, sampling_rate=self.sampling_rate
                    )

            # Compute the new Cut's duration after trimming the start and the end.
            new_duration = add_durations(
                track.cut.duration,
                -cut_offset,
                -cut_duration_decrease,
                sampling_rate=self.sampling_rate,
            )
            if new_duration <= 0:
                # Omit a Cut that is completely outside the time span of the new truncated MixedCut.
                continue

            new_tracks.append(
                MixTrack(
                    cut=track.cut.truncate(
                        offset=cut_offset,
                        duration=new_duration,
                        keep_excessive_supervisions=keep_excessive_supervisions,
                        preserve_id=preserve_id,
                        _supervisions_index=_supervisions_index,
                    ),
                    offset=track_offset,
                    snr=track.snr,
                )
            )

        # Edge case: no tracks with data left after truncation. This can happen if we truncated an offset region.
        # In this case, return a PaddingCut of the requested duration
        if len([t for t in new_tracks if not isinstance(t.cut, PaddingCut)]) == 0:
            return PaddingCut(
                id=self.id if preserve_id else str(uuid4()),
                duration=duration,
                sampling_rate=self.sampling_rate,
                feat_value=0.0,
                num_samples=compute_num_samples(duration, self.sampling_rate),
            )

        if len(new_tracks) == 1:
            # The truncation resulted in just a single cut - simply return it.
            return new_tracks[0].cut

        new_cut = MixedCut(
            id=self.id if preserve_id else str(uuid4()), tracks=new_tracks
        )

        # Final edge-case check. Scenario:
        # - some of the original MixedCut tracks had specified an SNR
        # - we truncated away the track that served as an SNR reference
        # - we are left only with PaddingCuts and MonoCuts that have specified SNR
        # Solution:
        # - find first non padding cut and reset its SNR to None (make it the new reference)
        if all(
            t.snr is not None or isinstance(t.cut, PaddingCut) for t in new_cut.tracks
        ):
            first_non_padding_track_idx = [
                idx
                for idx, t in enumerate(new_cut.tracks)
                if not isinstance(t.cut, PaddingCut)
            ][0]
            new_cut.tracks[first_non_padding_track_idx] = fastcopy(
                new_cut.tracks[first_non_padding_track_idx], snr=None
            )

        return new_cut

    def extend_by(
        self,
        *,
        duration: Seconds,
        direction: str = "both",
        preserve_id: bool = False,
        pad_silence: bool = True,
    ) -> "MixedCut":
        """
        This raises a ValueError since extending a MixedCut is not defined.

        :param duration: float (seconds), duration (in seconds) to extend the MixedCut.
        :param direction: string, 'left', 'right' or 'both'. Determines whether to extend on the left,
            right, or both sides. If 'both', extend on both sides by the duration specified in `duration`.
        :param preserve_id: bool. Should the extended cut keep the same ID or get a new, random one.
        :param pad_silence: bool. See usage in `lhotse.cut.MonoCut.extend_by`.
        :return: a new MixedCut instance.
        """
        raise ValueError("The extend_by() method is not defined for a MixedCut.")

    def pad(
        self,
        duration: Seconds = None,
        num_frames: int = None,
        num_samples: int = None,
        pad_feat_value: float = LOG_EPSILON,
        direction: str = "right",
        preserve_id: bool = False,
        pad_value_dict: Optional[Dict[str, Union[int, float]]] = None,
    ) -> Cut:
        """
        Return a new MixedCut, padded with zeros in the recording, and ``pad_feat_value`` in each feature bin.

        The user can choose to pad either to a specific `duration`; a specific number of frames `num_frames`;
        or a specific number of samples `num_samples`. The three arguments are mutually exclusive.

        :param duration: The cut's minimal duration after padding.
        :param num_frames: The cut's total number of frames after padding.
        :param num_samples: The cut's total number of samples after padding.
        :param pad_feat_value: A float value that's used for padding the features.
            By default we assume a log-energy floor of approx. -23 (1e-10 after exp).
        :param direction: string, 'left', 'right' or 'both'. Determines whether the padding is added before or after
            the cut.
        :param preserve_id: When ``True``, preserves the cut ID from before padding.
            Otherwise, generates a new random ID (default).
        :param pad_value_dict: Optional dict that specifies what value should be used
            for padding arrays in custom attributes.
        :return: a padded MixedCut if duration is greater than this cut's duration, otherwise ``self``.
        """
        from .set import pad

        return pad(
            self,
            duration=duration,
            num_frames=num_frames,
            num_samples=num_samples,
            pad_feat_value=pad_feat_value,
            direction=direction,
            preserve_id=preserve_id,
            pad_value_dict=pad_value_dict,
        )

    def resample(self, sampling_rate: int, affix_id: bool = False) -> "MixedCut":
        """
        Return a new ``MixedCut`` that will lazily resample the audio while reading it.
        This operation will drop the feature manifest, if attached.
        It does not affect the supervision.

        :param sampling_rate: The new sampling rate.
        :param affix_id: Should we modify the ID (useful if both versions of the same
            cut are going to be present in a single manifest).
        :return: a modified copy of the current ``MixedCut``.
        """
        assert self.has_recording, "Cannot resample a MixedCut without Recording."
        return MixedCut(
            id=f"{self.id}_rs{sampling_rate}" if affix_id else self.id,
            tracks=[
                fastcopy(t, cut=t.cut.resample(sampling_rate)) for t in self.tracks
            ],
        )

    def perturb_speed(self, factor: float, affix_id: bool = True) -> "MixedCut":
        """
        Return a new ``MixedCut`` that will lazily perturb the speed while loading audio.
        The ``num_samples``, ``start`` and ``duration`` fields of the underlying Cuts
        (and their Recordings and SupervisionSegments) are updated to reflect
        the shrinking/extending effect of speed.
        We are also updating the offsets of all underlying tracks.

        :param factor: The speed will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``MixedCut.id`` field
            by affixing it with "_sp{factor}".
        :return: a modified copy of the current ``MixedCut``.
        """
        # TODO(pzelasko): test most extensively for edge cases
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot perturb speed on a MixedCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to perturb speed on a MixedCut that references pre-computed features. "
                "The feature manifest(s) will be detached, as we do not support feature-domain "
                "speed perturbation."
            )
        return MixedCut(
            id=f"{self.id}_sp{factor}" if affix_id else self.id,
            tracks=[
                fastcopy(
                    track,
                    cut=track.cut.perturb_speed(factor=factor, affix_id=affix_id),
                    offset=round(
                        perturb_num_samples(
                            num_samples=compute_num_samples(
                                track.offset, self.sampling_rate
                            ),
                            factor=factor,
                        )
                        / self.sampling_rate,
                        ndigits=8,
                    ),
                )
                for track in self.tracks
            ],
        )

    def perturb_tempo(self, factor: float, affix_id: bool = True) -> "MixedCut":
        """
        Return a new ``MixedCut`` that will lazily perturb the tempo while loading audio.

        Compared to speed perturbation, tempo preserves pitch.
        The ``num_samples``, ``start`` and ``duration`` fields of the underlying Cuts
        (and their Recordings and SupervisionSegments) are updated to reflect
        the shrinking/extending effect of tempo.
        We are also updating the offsets of all underlying tracks.

        :param factor: The tempo will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``MixedCut.id`` field
            by affixing it with "_tp{factor}".
        :return: a modified copy of the current ``MixedCut``.
        """
        # TODO(pzelasko): test most extensively for edge cases
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot perturb tempo on a MixedCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to perturb tempo on a MixedCut that references pre-computed features. "
                "The feature manifest(s) will be detached, as we do not support feature-domain "
                "tempo perturbation."
            )
        return MixedCut(
            id=f"{self.id}_tp{factor}" if affix_id else self.id,
            tracks=[
                fastcopy(
                    track,
                    cut=track.cut.perturb_tempo(factor=factor, affix_id=affix_id),
                    offset=round(
                        perturb_num_samples(
                            num_samples=compute_num_samples(
                                track.offset, self.sampling_rate
                            ),
                            factor=factor,
                        )
                        / self.sampling_rate,
                        ndigits=8,
                    ),
                )
                for track in self.tracks
            ],
        )

    def perturb_volume(self, factor: float, affix_id: bool = True) -> "MixedCut":
        """
        Return a new ``MixedCut`` that will lazily perturb the volume while loading audio.
        Recordings of the underlying Cuts are updated to reflect volume change.

        :param factor: The volume will be adjusted this many times (e.g. factor=1.1 means 1.1x louder).
        :param affix_id: When true, we will modify the ``MixedCut.id`` field
            by affixing it with "_vp{factor}".
        :return: a modified copy of the current ``MixedCut``.
        """
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot perturb volume on a MixedCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to perturb volume on a MixedCut that references pre-computed features. "
                "The feature manifest(s) will be detached, as we do not support feature-domain "
                "volume perturbation."
            )
        return MixedCut(
            id=f"{self.id}_vp{factor}" if affix_id else self.id,
            tracks=[
                fastcopy(
                    track,
                    cut=track.cut.perturb_volume(factor=factor, affix_id=affix_id),
                )
                for track in self.tracks
            ],
        )

    def normalize_loudness(
        self, target: float, mix_first: bool = True, affix_id: bool = False
    ) -> "DataCut":
        """
        Return a new ``MixedCut`` that will lazily apply loudness normalization.

        :param target: The target loudness in dBFS.
        :param mix_first: If true, we will mix the underlying cuts before applying
            loudness normalization. If false, we cannot guarantee that the resulting
            cut will have the target loudness.
        :param affix_id: When true, we will modify the ``DataCut.id`` field
            by affixing it with "_ln{target}".
        :return: a modified copy of the current ``DataCut``.
        """
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot apply loudness normalization on a MixedCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to normalize loudness on a MixedCut that references pre-computed features. "
                "The feature manifest will be detached, as we do not support feature-domain "
                "loudness normalization."
            )
            self.features = None

        if mix_first:
            transforms = self.transforms.copy() if self.transforms is not None else []
            transforms.append(LoudnessNormalization(target=target))
            return fastcopy(
                self,
                id=f"{self.id}_ln{target}" if affix_id else self.id,
                transforms=transforms,
            )
        else:
            return MixedCut(
                id=f"{self.id}_ln{target}" if affix_id else self.id,
                tracks=[
                    fastcopy(
                        track,
                        cut=track.cut.normalize_loudness(
                            target=target, affix_id=affix_id
                        ),
                    )
                    for track in self.tracks
                ],
            )

    def reverb_rir(
        self,
        rir_recording: Optional["Recording"] = None,
        normalize_output: bool = True,
        early_only: bool = False,
        affix_id: bool = True,
        rir_channels: List[int] = [0],
        room_rng_seed: Optional[int] = None,
        source_rng_seed: Optional[int] = None,
        mix_first: bool = True,
    ) -> "MixedCut":
        """
        Return a new ``MixedCut`` that will convolve the audio with the provided impulse response.
        If no ``rir_recording`` is provided, we will generate an impulse response using a fast random
        generator (https://arxiv.org/abs/2208.04101).

        :param rir_recording: The impulse response to use for convolving.
        :param normalize_output: When true, output will be normalized to have energy as input.
        :param early_only: When true, only the early reflections (first 50 ms) will be used.
        :param affix_id: When true, we will modify the ``MixedCut.id`` field
            by affixing it with "_rvb".
        :param rir_channels: The channels of the impulse response to use. By default, first channel is used.
            If only one channel is specified, all tracks will be convolved with this channel. If a list
            is provided, it must contain as many channels as there are tracks such that each track will
            be convolved with one of the specified channels.
        :param room_rng_seed: Seed for the room configuration.
        :param source_rng_seed: Seed for the source position.
        :param mix_first: When true, the mixing will be done first before convolving with the RIR.
            This effectively means that all tracks will be convolved with the same RIR. If you
            are simulating multi-speaker mixtures, you should set this to False.
        :return: a modified copy of the current ``MixedCut``.
        """
        # Pre-conditions
        assert (
            self.has_recording
        ), "Cannot apply reverberation on a MixedCut without Recording."
        if self.has_features:
            logging.warning(
                "Attempting to reverberate a MixedCut that references pre-computed features. "
                "The feature manifest(s) will be detached, as we do not support feature-domain "
                "reverberation."
            )

        assert rir_recording is None or all(
            c < rir_recording.num_channels for c in rir_channels
        ), "Invalid channel index in `rir_channels`."

        assert len(rir_channels) == 1 or len(rir_channels) == len(
            self.tracks
        ), "Invalid number of channels in `rir_channels`, must be either 1 or equal to the number of tracks."

        # There are 2 ways to apply RIRs:
        # 1. Mix the tracks first, then apply RIRs. This is same as applying the same RIR
        #    to all tracks. It does not make sense if all tracks belong to different speakers,
        #    but it is useful for cases when we have a mixture of MonoCut and PaddingCut,
        #    and we want to apply the same RIR to all of them.
        # 2. Apply RIRs to each track separately. This is useful when we want to simulate
        #    different speakers in the same room.

        # First simulate the room config (will only be used if RIR is not provided)
        uuid4_str = str(uuid4())
        # The room RNG seed is based on the cut ID. This ensures that all tracks in the
        # mixed cut will have the same room configuration.
        if room_rng_seed is None:
            room_rng_seed = hash_str_to_int(uuid4_str + self.id)
        # The source RNG seed is based on the track ID. This ensures that each track
        # will have a different source position.
        source_rng_seeds = [source_rng_seed] * len(self.tracks)
        if source_rng_seed is None:
            source_rng_seeds = [
                hash_str_to_int(uuid4_str + track.cut.id) for track in self.tracks
            ]
            source_rng_seed = source_rng_seeds[0]

        # Apply same RIR to all tracks after mixing (default)
        if mix_first:
            if rir_recording is None:
                from lhotse.augmentation.utils import FastRandomRIRGenerator

                rir_generator = FastRandomRIRGenerator(
                    sr=self.sampling_rate,
                    room_seed=room_rng_seed,
                    source_seed=source_rng_seed,
                )
            else:
                rir_generator = None

            transforms = self.transforms.copy() if self.transforms is not None else []
            transforms.append(
                ReverbWithImpulseResponse(
                    rir=rir_recording,
                    normalize_output=normalize_output,
                    early_only=early_only,
                    rir_channels=rir_channels if rir_channels is not None else [0],
                    rir_generator=rir_generator,
                )
            )
            return fastcopy(
                self,
                id=f"{self.id}_rvb" if affix_id else self.id,
                transforms=transforms,
            )

        # Apply RIRs to each track separately. Note that we do not pass a `mix_first`
        # argument below since it is True by default.

        if len(rir_channels) == 1:
            rir_channels = rir_channels * len(self.tracks)

        return MixedCut(
            id=f"{self.id}_rvb" if affix_id else self.id,
            tracks=[
                fastcopy(
                    track,
                    cut=track.cut.reverb_rir(
                        rir_recording=rir_recording,
                        normalize_output=normalize_output,
                        early_only=early_only,
                        affix_id=affix_id,
                        rir_channels=[channel],
                        room_rng_seed=room_rng_seed,
                        source_rng_seed=seed,
                    ),
                )
                for track, channel, seed in zip(
                    self.tracks, rir_channels, source_rng_seeds
                )
            ],
        )

    @rich_exception_info
    def load_features(self, mixed: bool = True) -> Optional[np.ndarray]:
        """
        Loads the features of the source cuts and mixes them on-the-fly.

        :param mixed: when True (default), the features are mixed together (as defined in
            the mixing function for the extractor). This could result in either a 2D or 3D
            array. For example, if all underlying tracks are single-channel, the output
            will be a 2D array of shape (num_frames, num_features). If any of the tracks
            are multi-channel, the output may be a 3D array of shape (num_frames, num_features,
            num_channels).
        :return: A numpy ndarray with features and with shape ``(num_frames, num_features)``,
            or ``(num_tracks, num_frames, num_features)``
        """
        if not self.has_features:
            return None
        first_cut = self.tracks[0].cut

        # First, check for a simple scenario: just a single cut with padding.
        # When that is the case, we don't have to instantiate a feature extractor,
        # because we are not performing any actual mixing.
        # That makes life simpler for the users who have a custom feature extractor,
        # but don't actually care about feature-domain mixing; just want to pad.
        if mixed and all(isinstance(t.cut, PaddingCut) for t in self.tracks[1:]):
            padding_val = self.tracks[1].cut.feat_value
            first_cut_feats = first_cut.load_features()
            if first_cut_feats.ndim == 2:
                # 2D features
                feats = np.ones((self.num_frames, self.num_features)) * padding_val
            else:
                # 3D features
                feats = (
                    np.ones(
                        (self.num_frames, self.num_features, first_cut_feats.shape[-1])
                    )
                    * padding_val
                )
            feats[: first_cut.num_frames, ...] = first_cut_feats
            return feats

        # When there is more than one "regular" cut, we will perform an actual mix.

        # First, we have to make sure that the reference energy levels are appropriate.
        # They might not be if the first track is a padding track.
        reference_feats = None
        reference_energy = None
        reference_pos, reference_cut = [
            (idx, t.cut)
            for idx, t in enumerate(self.tracks)
            if not isinstance(t.cut, PaddingCut) and t.snr is None
        ][0]
        feature_extractor = create_default_feature_extractor(
            reference_cut.features.type
        )
        if first_cut.id != reference_cut.id:
            reference_feats = reference_cut.load_features()
            reference_energy = feature_extractor.compute_energy(reference_feats)

        # The mix itself.
        mixer = FeatureMixer(
            feature_extractor=create_default_feature_extractor(
                self._first_non_padding_cut.features.type
            ),
            base_feats=first_cut.load_features(),
            frame_shift=first_cut.frame_shift,
            reference_energy=reference_energy,
        )
        for pos, track in enumerate(self.tracks[1:], start=1):
            if pos == reference_pos and reference_feats is not None:
                feats = reference_feats  # manual caching to avoid duplicated I/O
            else:
                feats = track.cut.load_features()
            mixer.add_to_mix(
                feats=feats,
                snr=track.snr,
                offset=track.offset,
                sampling_rate=track.cut.sampling_rate,
            )

        if mixed:
            # Checking for some edge cases below.
            feats = mixer.mixed_feats
            # Note: The slicing below is a work-around for an edge-case
            #  when two cuts have durations that ended with 0.005 (e.g. 10.125 and 5.715)
            #  - then, the feature extractor "squeezed in" a last extra frame and the simple
            #  relationship between num_frames and duration we strived for is not true;
            #  i.e. the duration is 10.125 + 5.715 = 15.84, but the number of frames is
            #  1013 + 572 = 1585. If the frame_shift is 0.01, we have gained an extra 0.01s...
            if feats.shape[0] - self.num_frames == 1:
                feats = feats[: self.num_frames, :]
            # TODO(pzelasko): This can sometimes happen in a MixedCut with >= 5 different Cuts,
            #   with a regular MonoCut at the end, when the mix offsets are floats with a lot of decimals.
            #   For now we're duplicating the last frame to match the declared "num_frames" of this cut.
            if feats.shape[0] - self.num_frames == -1:
                feats = np.concatenate((feats, feats[-1:, :]), axis=0)
            assert feats.shape[0] == self.num_frames, (
                "Inconsistent number of frames in a MixedCut: please report "
                "this issue at https://github.com/lhotse-speech/lhotse/issues "
                "showing the output of print(cut) or str(cut) on which"
                "load_features() was called."
            )
            return feats
        else:
            return mixer.unmixed_feats

    @rich_exception_info
    def load_audio(
        self, mixed: bool = True, mono_downmix: bool = False
    ) -> Optional[np.ndarray]:
        """
        Loads the audios of the source cuts and mix them on-the-fly.

        :param mixed: When True (default), returns a mix of the underlying tracks. This will
            return a numpy array with shape ``(num_channels, num_samples)``, where ``num_channels``
            is determined by the ``num_channels`` property of the MixedCut. Otherwise returns a
            numpy array with the number of channels equal to the total number of channels
            across all tracks in the MixedCut. For example, if it contains a MultiCut with 2
            channels and a MonoCut with 1 channel, the returned array will have shape
            ``(3, num_samples)``.
        :param mono_downmix: If the MixedCut contains > 1 channels (for e.g. when one of its tracks
            is a MultiCut), this parameter controls whether the returned array will be down-mixed
            to a single channel. This down-mixing is done by summing the channels together.
        :return: A numpy ndarray with audio samples and with shape ``(num_channels, num_samples)``
        """

        if not self.has_recording:
            return None
        first_cut = self.tracks[0].cut

        # First, we have to make sure that the reference energy levels are appropriate.
        # They might not be if the first track is a padding track.
        reference_audio = None
        reference_energy = None
        reference_pos, reference_cut = [
            (idx, t.cut)
            for idx, t in enumerate(self.tracks)
            if not isinstance(t.cut, PaddingCut) and t.snr is None
        ][0]
        if first_cut.id != reference_cut.id:
            reference_audio = reference_cut.load_audio()
            reference_energy = audio_energy(reference_audio)

        mixer = AudioMixer(
            self.tracks[0].cut.load_audio(),
            sampling_rate=self.tracks[0].cut.sampling_rate,
            reference_energy=reference_energy,
            base_offset=self.tracks[0].offset,
        )

        for pos, track in enumerate(self.tracks[1:], start=1):
            if pos == reference_pos and reference_audio is not None:
                audio = reference_audio  # manual caching to avoid duplicated I/O
            else:
                audio = track.cut.load_audio()
            mixer.add_to_mix(
                audio=audio,
                snr=track.snr,
                offset=track.offset,
            )

        # Flattening a MixedCut without MultiCut tracks has no effect
        mono_downmix = mono_downmix and any(
            track.type == "MultiCut" for track in self.tracks
        )

        # Flattening a MixedCut without mixed=True has no effect
        mono_downmix = mono_downmix and mixed

        if mixed:
            # Off-by-one errors can happen during mixing due to imperfect float arithmetic and rounding;
            # we will fix them on-the-fly so that the manifest does not lie about the num_samples.
            audio = mixer.mixed_mono_audio if mono_downmix else mixer.mixed_audio
            tol_samples = compute_num_samples(
                get_audio_duration_mismatch_tolerance(),
                sampling_rate=self.sampling_rate,
            )
            num_samples_diff = audio.shape[1] - self.num_samples
            if 0 < num_samples_diff < tol_samples:
                audio = audio[:, : self.num_samples]
            if -tol_samples < num_samples_diff < 0:
                audio = np.pad(audio, [(0, 0), (0, -num_samples_diff)], mode="reflect")
            assert audio.shape[1] == self.num_samples, (
                f"Inconsistent number of samples in a MixedCut. Expected {self.num_samples} "
                f"but the output of mix has {audio.shape[1]}. Please report "
                f"this issue at https://github.com/lhotse-speech/lhotse/issues "
                f"showing the cut below. MixedCut:\n{self}"
            )

            # We'll apply the transforms now (if any).
            transforms = [
                tnfm
                if isinstance(tnfm, AudioTransform)
                else AudioTransform.from_dict(tnfm)
                for tnfm in self.transforms or []
            ]
            for tfn in transforms:
                audio = tfn(audio, self.sampling_rate)
        else:
            audio = mixer.unmixed_audio

        return audio

    @property
    def video(self) -> Optional[VideoInfo]:
        if self.has_video:
            v = self._first_non_padding_cut.video
            return v.copy_with(num_frames=compute_num_samples(self.duration, v.fps))
        return None

    @rich_exception_info
    def load_video(
        self,
        with_audio: bool = True,
        mixed: bool = True,
        mono_downmix: bool = False,
    ) -> Optional[Tuple[torch.Tensor, Optional[torch.Tensor]]]:
        if not self.has_video:
            return None

        mixer = VideoMixer(
            self.tracks[0].cut.load_video(with_audio=False)[0],
            fps=self.video.fps,
            base_offset=self.tracks[0].offset,
        )
        for pos, track in enumerate(self.tracks[1:], start=1):
            mixer.add_to_mix(
                video=track.cut.load_video(with_audio=False)[0],
                offset=track.offset,
            )
        video = mixer.mixed_video

        if with_audio:
            # For now load audio separately to re-use the complex logic of load_audio
            # This means the same video file is potentially opened twice, but given the
            # cost of video decoding, the extra file open cost could be negligible.
            audio = self.load_audio(mixed=mixed, mono_downmix=mono_downmix)
        return video, torch.from_numpy(audio)

    def plot_tracks_features(self):
        """
        Display the feature matrix as an image. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(len(self.tracks))
        features = self.load_features(mixed=False)
        fmin, fmax = features.min(), features.max()
        for idx, ax in enumerate(axes):
            ax.imshow(np.flip(features[idx].transpose(1, 0), 0), vmin=fmin, vmax=fmax)
        return axes

    def plot_tracks_audio(self):
        """
        Display plots of the individual tracks' waveforms. Requires matplotlib to be installed.
        """
        import matplotlib.pyplot as plt

        audio = self.load_audio(mixed=False)
        fig, axes = plt.subplots(len(self.tracks), sharex=False, sharey=True)
        for idx, (track, ax) in enumerate(zip(self.tracks, axes)):
            samples = audio[idx].squeeze(0)
            ax.plot(np.linspace(0, self.duration, len(samples)), samples)
            for supervision in track.cut.supervisions:
                supervision = supervision.trim(track.cut.duration)
                ax.axvspan(
                    track.offset + supervision.start,
                    track.offset + supervision.end,
                    color="green",
                    alpha=0.1,
                )
        return axes

    def drop_features(self) -> "MixedCut":
        """Return a copy of the current :class:`MixedCut`, detached from ``features``."""
        assert (
            self.has_recording
        ), f"Cannot detach features from a MixedCut with no Recording (cut ID = {self.id})."
        return fastcopy(
            self, tracks=[fastcopy(t, cut=t.cut.drop_features()) for t in self.tracks]
        )

    def drop_recording(self) -> "MixedCut":
        """Return a copy of the current :class:`.MixedCut`, detached from ``recording``."""
        assert (
            self.has_features
        ), f"Cannot detach recording from a MixedCut with no Features (cut ID = {self.id})."
        return fastcopy(
            self, tracks=[fastcopy(t, cut=t.cut.drop_recording()) for t in self.tracks]
        )

    def drop_supervisions(self) -> "MixedCut":
        """Return a copy of the current :class:`.MixedCut`, detached from ``supervisions``."""
        return fastcopy(
            self,
            tracks=[fastcopy(t, cut=t.cut.drop_supervisions()) for t in self.tracks],
        )

    def drop_alignments(self) -> "MixedCut":
        """Return a copy of the current :class:`.MixedCut`, detached from ``supervisions``."""
        return fastcopy(
            self,
            tracks=[fastcopy(t, cut=t.cut.drop_alignments()) for t in self.tracks],
        )

    def drop_in_memory_data(self) -> "MixedCut":
        """Return a copy of the current :class:`MixedCut`, which doesn't contain any in-memory data."""
        return fastcopy(
            self,
            tracks=[fastcopy(t, cut=t.cut.drop_in_memory_data()) for t in self.tracks],
        )

    def compute_and_store_features(
        self,
        extractor: FeatureExtractor,
        storage: FeaturesWriter,
        augment_fn: Optional[AugmentFn] = None,
        mix_eagerly: bool = True,
    ) -> DataCut:
        """
        Compute the features from this cut, store them on disk, and create a new `MonoCut` object with the
        feature manifest attached. This cut has to be able to load audio.

        :param extractor: a ``FeatureExtractor`` instance used to compute the features.
        :param storage: a ``FeaturesWriter`` instance used to store the features.
        :param augment_fn: an optional callable used for audio augmentation.
        :param mix_eagerly: when False, extract and store the features for each track separately,
            and mix them dynamically when loading the features.
            When True, mix the audio first and store the mixed features, returning a new ``MonoCut`` instance
            with the same ID. The returned ``MonoCut`` will not have a ``Recording`` attached.
        :return: a new ``MonoCut`` instance if ``mix_eagerly`` is True, or returns ``self``
            with each of the tracks containing the ``Features`` manifests.
        """
        if mix_eagerly:
            from .mono import MonoCut

            features_info = extractor.extract_from_samples_and_store(
                samples=self.load_audio(),
                storage=storage,
                sampling_rate=self.sampling_rate,
                offset=0,
                channel=0,
                augment_fn=augment_fn,
            )
            features_info.recording_id = self.id
            return MonoCut(
                id=self.id,
                start=0,
                duration=self.duration,
                channel=0,
                supervisions=[
                    fastcopy(s, recording_id=self.id) for s in self.supervisions
                ],
                features=features_info,
                recording=None,
                custom=self.custom if hasattr(self, "custom") else None,
            )
        else:  # mix lazily
            new_tracks = [
                MixTrack(
                    cut=track.cut.compute_and_store_features(
                        extractor=extractor,
                        storage=storage,
                        augment_fn=augment_fn,
                    ),
                    offset=track.offset,
                    snr=track.snr,
                )
                for track in self.tracks
            ]
            return MixedCut(id=self.id, tracks=new_tracks)

    def fill_supervision(
        self, add_empty: bool = True, shrink_ok: bool = False
    ) -> "MixedCut":
        """
        Fills the whole duration of a cut with a supervision segment.

        If the cut has one supervision, its start is set to 0 and duration is set to ``cut.duration``.
        Note: this may either expand a supervision that was shorter than a cut, or shrink a supervision
        that exceeds the cut.

        If there are no supervisions, we will add an empty one when ``add_empty==True``, otherwise
        we won't change anything.

        If there are two or more supervisions, we will raise an exception.

        .. note:: For :class:`.MixedCut`, we expect that only one track contains a supervision.
            That supervision will be expanded to cover the full MixedCut's duration.

        :param add_empty: should we add an empty supervision with identical time bounds as the cut.
        :param shrink_ok: should we raise an error if a supervision would be shrank as a result
            of calling this method.
        """
        n_sups = len(self.supervisions)
        if n_sups == 0:
            if not add_empty:
                return self
            first_non_padding_idx = [
                idx for idx, t in enumerate(self.tracks) if isinstance(t.cut, DataCut)
            ][0]
            new_tracks = [
                fastcopy(
                    t,
                    cut=fastcopy(
                        t.cut,
                        supervisions=[
                            SupervisionSegment(
                                id=self.id,
                                recording_id=t.cut.recording_id,
                                start=-t.offset,
                                duration=self.duration,
                                channel=-1,
                            )
                        ],
                    ),
                )
                if idx == first_non_padding_idx
                else t
                for idx, t in enumerate(self.tracks)
            ]
        else:
            assert (
                n_sups == 1
            ), f"Cannot expand more than one supervision (found {len(self.supervisions)}."
            new_tracks = []
            for t in self.tracks:
                if len(t.cut.supervisions) == 0:
                    new_tracks.append(t)
                else:
                    sup = t.cut.supervisions[0]
                    if not shrink_ok and (
                        sup.start < -t.offset or sup.end > self.duration
                    ):
                        raise ValueError(
                            f"Cannot shrink supervision (start={sup.start}, end={sup.end}) to cut "
                            f"(start=0, duration={t.cut.duration}) because the argument `shrink_ok` is `False`. "
                            f"Note: this check prevents accidental data loss for speech recognition, "
                            f"as supervision exceeding a cut indicates there might be some spoken content "
                            f"beyond cuts start or end (an ASR model would be trained to predict more text than "
                            f"spoken in the audio). If this is okay, set `shrink_ok` to `True`."
                        )
                    new_tracks.append(
                        fastcopy(
                            t,
                            cut=fastcopy(
                                t.cut,
                                supervisions=[
                                    fastcopy(
                                        sup, start=-t.offset, duration=self.duration
                                    )
                                ],
                            ),
                        )
                    )
        return fastcopy(self, tracks=new_tracks)

    def map_supervisions(
        self, transform_fn: Callable[[SupervisionSegment], SupervisionSegment]
    ) -> Cut:
        """
        Modify the SupervisionSegments by `transform_fn` of this MixedCut.

        :param transform_fn: a function that modifies a supervision as an argument.
        :return: a modified MixedCut.
        """
        new_mixed_cut = fastcopy(self)
        for track in new_mixed_cut.tracks:
            if isinstance(track.cut, PaddingCut):
                continue
            track.cut.supervisions = [
                segment.map(transform_fn) for segment in track.cut.supervisions
            ]
        return new_mixed_cut

    def merge_supervisions(
        self,
        merge_policy: str = "delimiter",
        custom_merge_fn: Optional[Callable[[str, Iterable[Any]], Any]] = None,
    ) -> "MixedCut":
        """
        Return a copy of the cut that has all of its supervisions merged into
        a single segment.

        The new start is the start of the earliest superivion, and the new duration
        is a minimum spanning duration for all the supervisions. The text fields are
        concatenated with a whitespace.

        .. note:: If you're using individual tracks of a mixed cut, note that this transform
             drops all the supervisions in individual tracks and assigns the merged supervision
             in the first :class:`.DataCut` found in ``self.tracks``.

        :param merge_policy: one of "keep_first" or "delimiter". If "keep_first", we
            keep only the first segment's field value, otherwise all string fields
            (including IDs) are prefixed with "cat#" and concatenated with a hash symbol "#".
            This is also applied to ``custom`` fields. Fields with a ``None`` value are omitted.
        :param custom_merge_fn: a function that will be called to merge custom fields values.
            We expect ``custom_merge_fn`` to handle all possible custom keys.
            When not provided, we will treat all custom values as strings.
            It will be called roughly like:
            ``custom_merge_fn(custom_key, [s.custom[custom_key] for s in sups])``
        """
        merge_func_ = partial(
            merge_items_with_delimiter,
            delimiter="#",
            return_first=(merge_policy == "keep_first"),
        )

        # "m" stands for merged in variable names below

        if custom_merge_fn is not None:
            # Merge custom fields with the user-provided function.
            merge_custom = custom_merge_fn
        else:
            # Merge the string representations of custom fields.
            merge_custom = lambda k, vs: merge_func_(map(str, vs))

        sups = sorted(self.supervisions, key=lambda s: s.start)

        if len(sups) <= 1:
            return self

        # the sampling rate is arbitrary, ensures there are no float precision errors
        mstart = sups[0].start
        mend = sups[-1].end
        mduration = add_durations(mend, -mstart, sampling_rate=self.sampling_rate)

        custom_keys = set(
            k for s in sups if s.custom is not None for k in s.custom.keys()
        )
        alignment_keys = set(
            k for s in sups if s.alignment is not None for k in s.alignment.keys()
        )

        if any(overlaps(s1, s2) for s1, s2 in zip(sups, sups[1:])) and any(
            s.text is not None for s in sups
        ):
            warnings.warn(
                "You are merging overlapping supervisions that have text transcripts. "
                "The result is likely to be unusable if you are going to train speech "
                f"recognition models (cut id: {self.id})."
            )

        msup = SupervisionSegment(
            id=merge_func_(s.id for s in sups),
            # Make merged recording_id is a mix of recording_ids.
            recording_id=merge_func_(s.recording_id for s in sups),
            start=mstart,
            duration=mduration,
            # Hardcode -1 to indicate no specific channel, as the supervisions might have
            # come from different channels in their original recordings.
            channel=-1,
            text=" ".join(s.text for s in sups if s.text),
            speaker=merge_func_(s.speaker for s in sups if s.speaker),
            language=merge_func_(s.language for s in sups if s.language),
            gender=merge_func_(s.gender for s in sups if s.gender),
            custom={
                k: merge_custom(
                    k,
                    (
                        s.custom[k]
                        for s in sups
                        if s.custom is not None and k in s.custom
                    ),
                )
                for k in custom_keys
            },
            alignment={
                # Concatenate the lists of alignment units.
                k: reduce(
                    add,
                    (
                        s.alignment[k]
                        for s in sups
                        if s.alignment is not None and k in s.alignment
                    ),
                )
                for k in alignment_keys
            },
        )

        new_cut = self.drop_supervisions()
        new_cut._first_non_padding_cut.supervisions = [msup]
        return new_cut

    def filter_supervisions(
        self, predicate: Callable[[SupervisionSegment], bool]
    ) -> Cut:
        """
        Modify cut to store only supervisions accepted by `predicate`

        Example:
            >>> cut = cut.filter_supervisions(lambda s: s.id in supervision_ids)
            >>> cut = cut.filter_supervisions(lambda s: s.duration < 5.0)
            >>> cut = cut.filter_supervisions(lambda s: s.text is not None)

        :param predicate: A callable that accepts `SupervisionSegment` and returns bool
        :return: a modified MixedCut
        """
        new_mixed_cut = fastcopy(
            self,
            tracks=[
                fastcopy(track, cut=track.cut.filter_supervisions(predicate))
                for track in self.tracks
            ],
        )
        return new_mixed_cut

    @staticmethod
    def from_dict(data: dict) -> "MixedCut":
        if "type" in data:
            data.pop("type")
        transforms = None
        if "transforms" in data:
            transforms = [AudioTransform.from_dict(t) for t in data["transforms"]]
        return MixedCut(
            id=data["id"],
            tracks=[MixTrack.from_dict(track) for track in data["tracks"]],
            transforms=transforms,
        )

    def with_features_path_prefix(self, path: Pathlike) -> "MixedCut":
        if not self.has_features:
            return self
        return MixedCut(
            id=self.id,
            tracks=[
                fastcopy(t, cut=t.cut.with_features_path_prefix(path))
                for t in self.tracks
            ],
        )

    def with_recording_path_prefix(self, path: Pathlike) -> "MixedCut":
        if not self.has_recording:
            return self
        return MixedCut(
            id=self.id,
            tracks=[
                fastcopy(t, cut=t.cut.with_recording_path_prefix(path))
                for t in self.tracks
            ],
        )

    @property
    def first_non_padding_cut(self) -> DataCut:
        return self._first_non_padding_track.cut

    @property
    def first_non_padding_track(self) -> MixTrack:
        return [t for t in self.tracks if not isinstance(t.cut, PaddingCut)][0]

    # Note: the private properties below are kept for backward compatibility.

    @property
    def _first_non_padding_cut(self) -> DataCut:
        return self.first_non_padding_cut

    @property
    def _first_non_padding_track(self) -> MixTrack:
        return self.first_non_padding_track
